"""
------------------------------------------
File name: barotropic_field.py
Author: Clare Huang
"""
from math import pi

import numpy as np

from falwa import basis


class BarotropicField(object):

    """
    An object that deals with barotropic (2D) wind and/or PV fields

    Parameters
    ----------
        xlon : np.array
            Longitude array in degree with dimension = nlon.

        ylat : np.array
            Latitude array in degree with dimension = nlat.

        area : np.ndarray
            Differential area at each lon-lat grid points with dimension (nlat,nlon). If 'area=None': it will be initiated as area of uniform grid (in degree) on a spherical surface. Dimension = [nlat, nlon]

        dphi : np.array
            Differential length element along the lat grid with dimension = nlat.

        pv_field : np.ndarray
            Absolute vorticity field with dimension [nlat x nlon]. If 'pv_field=None': pv_field is expected to be computed with u,v,t field.


    Examples
    --------
    :doc:`Example Notebook <notebooks/Example_qgpv>`

    >>> barofield1 = BarotropicField(xlon, ylat, pv_field=abs_vorticity)

    """

    def __init__(self, xlon, ylat, pv_field, area=None, dphi=None,
                 n_partitions=None, planet_radius=6.378e+6):

        """
        Create a BarotropicField object.

        Parameters
        ----------
            xlon : np.array
                Longitude array in degree with dimension = nlon.

            ylat : np.array
                Latitude array in degree with dimension = nlat.

            area : np.ndarray
                Differential area at each lon-lat grid points with dimension (nlat,nlon). If 'area=None': it will be initiated as area of uniform grid (in degree) on a spherical surface. Dimension = [nlat, nlon]

            dphi : np.array
                Differential length element along the lat grid with dimension = nlat.

            pv_field : np.ndarray
                Absolute vorticity field with dimension = [nlat, nlon].
                If none, pv_field is expected to be computed with u,v,t field.

        """

        self.xlon = xlon
        self.ylat = ylat
        self.clat = np.abs(np.cos(np.deg2rad(ylat)))
        self.nlon = xlon.size
        self.nlat = ylat.size
        self.planet_radius = planet_radius
        if dphi is None:
            self.dphi = pi/(self.nlat-1) * np.ones((self.nlat))
        else:
            self.dphi = dphi

        if area is None:
            self.area = 2. * pi * planet_radius ** 2 * \
                        (np.cos(ylat[:, np.newaxis] * pi/180.) * self.dphi[:, np.newaxis])\
                        / float(self.nlon)*np.ones((self.nlat, self.nlon))
        else:
            self.area = area

        self.pv_field = pv_field

        if n_partitions is None:
            self.n_partitions = self.nlat
        else:
            self.n_partitions = n_partitions

        # Quantities that are computed with the methods below
        self._eqvlat = None
        self._lwa = None

    def _compute_eqvlat(self):
        """
        Private function. Compute equivalent latitude if it has not been computed yet.
        """
        self.eqvlat, _ = basis.eqvlat_fawa(
            self.ylat, self.pv_field, self.area, self.n_partitions,
            planet_radius=self.planet_radius, output_fawa=False
        )
        return self.eqvlat

    def _compute_lwa(self):
        """
        Private function. Compute equivalent latitude if it has not been computed yet.
        """
        eqvlat = self.equivalent_latitudes
        if self._lwa is None:
            self._lwa, dummy = basis.lwa(
                self.nlon, self.nlat, self.pv_field, eqvlat,
                self.planet_radius * self.clat * self.dphi
            )
        return self.lwa

    @property
    def equivalent_latitudes(self):
        """
        Return the computd quivalent latitude with the *pv_field* stored in the object.

        Return
        ----------
        An numpy array with dimension (nlat) of equivalent latitude array.

        Example
        ----------
        >>> barofield1 = BarotropicField(xlon, ylat, pv_field=abs_vorticity)
        >>> eqv_lat = barofield1.equivalent_latitudes

        """
        if self._eqvlat is None:
            return self._compute_eqvlat()
        return self._eqvlat

    @property
    def lwa(self):

        """
        Compute the finite-amplitude local wave activity based on the *equivalent_latitudes* and the *pv_field* stored in the object.

        Return
        ----------
        An 2-D numpy array with dimension [nlat,nlon] of local wave activity values.

        Example
        ----------
        >>> barofield1 = BarotropicField(xlon, ylat, pv_field=abs_vorticity)
        >>> lwa = barofield1.lwa

        """
        if self._lwa is None:
            return self._compute_lwa()
        return self._lwa



"""
------------------------------------------
File name: __init__.py
Author: Clare Huang, Christopher Polster
"""

__version__ = "1.2.1"
from .interpolate_fields import interpolate_fields
from .interpolate_fields_direct_inv import interpolate_fields_direct_inv
from .compute_qref_and_fawa_first import compute_qref_and_fawa_first
from .matrix_b4_inversion import matrix_b4_inversion
from .matrix_after_inversion import matrix_after_inversion
from .upward_sweep import upward_sweep
from .compute_reference_states import compute_reference_states
from .compute_flux_dirinv import compute_flux_dirinv_nshem
from .compute_lwa_and_barotropic_fluxes import compute_lwa_and_barotropic_fluxes



import pytest
import os


@pytest.fixture(scope="session")
def test_data_dir() -> str:
    """
    Return the path of directory of data storage for unit tests
    """
    return os.path.dirname(os.path.abspath(__file__)) + "/data"





import warnings


def input_jk_output_index(j,k,kmax):
    return j*(kmax) + k


def extrap1d(interpolator):

    xs = interpolator.x
    ys = interpolator.y

    def pointwise(x):
        if x < xs[0]:
            return ys[0]+(x-xs[0])*(ys[1]-ys[0])/(xs[1]-xs[0])
        elif x > xs[-1]:
            return ys[-1]+(x-xs[-1])*(ys[-1]-ys[-2])/(xs[-1]-xs[-2])
        else:
            return interpolator(x)

    def ufunclike(xs):
        from scipy import array
        return array(map(pointwise, array(xs)))

    return ufunclike


def solve_uref_both_bc(tstamp, zmum, FAWA_cos, ylat, ephalf2, Delta_PT,
                       zm_PT, Input_B0, Input_B1, use_real_Data=True,
                       plot_all_ref_quan=False):
    """
    Compute equivalent latitude and wave activity on a barotropic sphere.

    Parameters
    ----------
    tstamp : string
        Time stamp of the snapshot of the field.
    znum : ndarray
        Zonal mean wind.
    FAWA_cos : ndarray
        Zonal mean finite-amplitude wave activity.
    ylat : sequence or array_like
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    ephalf2 : ndarray
        Epsilon in Nakamura and Solomon (2010).
    Delta_PT : ndarray
        \Delta \Theta in Nakamura and Solomon (2010); upper-boundary conditions.
    zm_PT : ndarray
        Zonal mean potential temperature.
    Input_B0 : sequence or array_like
        Zonal-mean surface wave activity for the lowest layer (k=0). Part of the lower-boundary condition.
    Input_B1 : sequence or array_like
        Zonal-mean surface wave activity for the second lowest layer (k=1). Part of the lower-boundary condition.
    use_real_Data : boolean
        Whether to use input data to compute the reference states. By detault True. If false, randomly generated arrays will be used.
    plot_all_ref_quan : boolean
        Whether to plot the solved reference states using matplotlib library. By default False. For debugging.



    Returns
    -------
    u_MassCorr_regular_noslip : ndarray
        2-d numpy array of mass correction \Delta u in NS10 with no-slip lower boundary conditions; dimension = (kmax,nlat).
    u_Ref_regular_noslip : ndarray
        2-d numpy array of zonal wind reference state u_ref in NS10 with no-slip lower boundary conditions; dimension = (kmax,nlat).
    T_MassCorr_regular_noslip : ndarray
        2-d numpy array of adjustment in reference temperature \Delta T in NS10 with no-slip lower boundary conditions; dimension = (kmax,nlat).
    T_Ref_regular_noslip : ndarray
        2-d numpy array of adjustment in reference temperature T_ref in NS10 with no-slip lower boundary conditions; dimension = (kmax,nlat).
    u_MassCorr_regular_adiab : ndarray
        2-d numpy array of mass correction \Delta u in NS10 with adiabatic lower boundary conditions; dimension = (kmax,nlat).
    u_Ref_regular_adiab : ndarray
        2-d numpy array of zonal wind reference state u_ref in NS10 with adiabatic lower boundary conditions; dimension = (kmax,nlat).
    T_MassCorr_regular_adiab : ndarray
        2-d numpy array of adjustment in reference temperature \Delta T in NS10 with adiabatic lower boundary conditions; dimension = (kmax,nlat).
    T_Ref_regular_adiab : ndarray
        2-d numpy array of adjustment in reference temperature T_ref in NS10 with adiabatic lower boundary conditions; dimension = (kmax,nlat).

    """

    warnings.warn(
        """The function solve_uref_both_bc is no longer maintained and will be deprecated.
        Please use the class oopinterface.QGField for the computation of reference states.
        """,
        DeprecationWarning
    )
    # zm_PT = zonal mean potential temperature

    # Import necessary modules
    from math import pi, exp
    from scipy import interpolate
    from scipy.sparse import csc_matrix
    from scipy.sparse.linalg import spsolve
    from copy import copy
    import numpy as np
    import itertools
    if plot_all_ref_quan:
        import matplotlib.pyplot as plt

    # === Parameters (should be input externally. To be modified) ===
    dz = 1000.          # vertical z spacing (m)
    aa = 6378000.     # planetary radius
    r0 = 287.           # gas constant
    hh = 7000.          # scale height
    cp = 1004.          # specific heat
    rkappa = r0/cp
    om = 7.29e-5          # angular velocity of the earth

    # === These changes with input variables' dimensions ===
    nlat = FAWA_cos.shape[-1]
    jmax1 = nlat//4
    dm = 1./float(jmax1+1)  # gaussian latitude spacing
    gl = np.array([(j+1)*dm for j in range(jmax1)]) # This is sin / mu
    gl_2 = np.array([j*dm for j in range(jmax1+2)]) # This is sin / mu
    cosl = np.sqrt(1.-gl**2)
    #cosl_2 = np.sqrt(1.-gl_2**2)
    alat = np.arcsin(gl)*180./pi
    alat_2 = np.arcsin(gl_2)*180./pi
    dmdz = (dm/dz)

    # **** Get from input these parameters ****
    kmax = FAWA_cos.shape[0]
    #height = np.array([i for i in range(kmax)]) # in [km]

    # **** Initialize Coefficients ****
    c_a = np.zeros((jmax1, kmax))
    c_b = np.zeros((jmax1, kmax))
    c_c = np.zeros((jmax1, kmax))
    c_d = np.zeros((jmax1, kmax))
    c_e = np.zeros((jmax1, kmax))
    c_f = np.zeros((jmax1, kmax))

    # --- Initialize interpolated variables ---
    zmu1 = np.zeros((jmax1, kmax))
    cx1 = np.zeros((jmax1, kmax))
    cor1 = np.zeros((jmax1, kmax))
    ephalf = np.zeros((jmax1, kmax))
    Delta_PT1 = np.zeros((jmax1+2))
    zm_PT1 = np.zeros((jmax1, kmax))
    Input_B0_1 = np.zeros((jmax1+2))
    Input_B1_1 = np.zeros((jmax1+2))

    # --- Define Epsilon as a function of y and z ---

    # **** Interpolate to gaussian latitude ****
    if use_real_Data:
        # print 'use_real_Data'
        for vv1,vvm in zip([zmu1,cx1,zm_PT1] , [zmum,FAWA_cos,zm_PT]):
            f_toGaussian = interpolate.interp1d(ylat[:],vvm[:,:].T,axis=0, kind='linear')    #[jmax x kmax]
            vv1[:,:] = f_toGaussian(alat[:])
            #vv1[:,:] = vvm[:,:]
            #vv1[-1,:] = vvm[:,-1]

        # --- Interpolation of ephalf ---
        f_ep_toGaussian = interpolate.interp1d(ylat[:],ephalf2[:,:].T,axis=0, kind='linear')    #[jmax x kmax]
        ephalf[:,:] = f_ep_toGaussian(alat[:])

		# --- Interpolation of Delta_PT ---
        #f_DT_toGaussian = extrap1d( interpolate.interp1d(ylat[:],Delta_PT[:], kind='linear') )    # This is txt in Noboru's code
        f_DT_toGaussian = interpolate.interp1d(ylat[:],Delta_PT[:],
                                               kind='linear',fill_value='extrapolate')
        Delta_PT1[:] = f_DT_toGaussian(alat_2[:])

		# --- Interpolation of Input_B0_1 ---
        #f_B0_toGaussian = extrap1d( interpolate.interp1d(ylat[:],Input_B0[:], kind='linear') )    # This is txt in Noboru's code
        f_B0_toGaussian = interpolate.interp1d(ylat[:],Input_B0[:],
                                               kind='linear',fill_value='extrapolate')    # This is txt in Noboru's code
        Input_B0_1[:] = f_B0_toGaussian(alat_2[:])

		# --- Interpolation of Input_B1_1 ---
        # f_B1_toGaussian = extrap1d( interpolate.interp1d(ylat[:],Input_B1[:], kind='linear') )     # This is txt in Noboru's code
        f_B1_toGaussian = interpolate.interp1d(ylat[:],Input_B1[:],
                                               kind='linear',fill_value='extrapolate')    # This is txt in Noboru's code
        Input_B1_1[:] = f_B1_toGaussian(alat_2[:])

    else:
        # Use random matrix here just to test!
        zmu1 = np.random.rand(jmax1, kmax)+np.ones((jmax1, kmax))*1.e-8
        cx1 = np.random.rand(jmax1, kmax)+np.ones((jmax1, kmax))*1.e-8
        #cor1 = np.random.rand(jmax1, kmax)+np.ones((jmax1, kmax))*1.e-8


    # --- Added on Aug 1, 2016 ---
    cor1 = 2.*om*gl[:,np.newaxis] * np.ones((jmax1, kmax))
    #cor1[0] = cor1[1]*0.5

    # OLD: qxx0 = -cx1*cosl[:,np.newaxis]/cor1     #qxx0 = np.empty((jmax1, kmax))
    qxx0 = -cx1/cor1 # Input of LWA has cosine.
    c_f[0,:] = qxx0[1,:] - 2*qxx0[0,:]
    c_f[-1,:] = qxx0[-2,:] - 2*qxx0[-1,:]
    c_f[1:-1,:] = qxx0[:-2,:] + qxx0[2:,:] - 2*qxx0[1:-1,:]
    #c_f[:,0] = 0.0

    # --- Aug 9: Lower Adiabatic boundary conditions ---
    Input_dB0 = np.zeros((jmax1))
    Input_dB1 = np.zeros((jmax1))
    uz1 = np.zeros((jmax1))

    # prefac = - r0 * cosl[1:-1]**2 * dz / (cor1[1:-1,-2]**2 * aa**2 * hh * dm**2) * exp(-rkappa*(kmax-2.)/7.)

    # OLD: Input_dB0[:] = Input_B0_1[:-2]*cosl_2[:-2] + Input_B0_1[2:]*cosl_2[2:] - 2*Input_B0_1[1:-1]*cosl_2[1:-1]
    Input_dB0[:] = Input_B0_1[:-2] + Input_B0_1[2:] - 2*Input_B0_1[1:-1]

    # OLD: Input_dB1[:] = Input_B1_1[:-2]*cosl_2[:-2] + Input_B1_1[2:]*cosl_2[2:] - 2*Input_B1_1[1:-1]*cosl_2[1:-1]
    Input_dB1[:] = Input_B1_1[:-2] + Input_B1_1[2:] - 2*Input_B1_1[1:-1]

    # This is supposed to be correct but gave weird results.
    uz1[:] = - r0 * cosl[:]**2 * Input_dB1[:] * 2*dz / (cor1[:,1]**2 * aa**2 * hh * dm**2) * exp(-rkappa*(1.)/7.) \
    - r0 * cosl[:]**2 * Input_dB0[:] * 2*dz / (cor1[:,0]**2 * aa**2 * hh * dm**2) * exp(-rkappa*(0.)/7.)

    #  **** Upper Boundary Condition (Come back later) ****
    uz2 = np.zeros((jmax1))
    dDelta_PT1 = (Delta_PT1[2:]-Delta_PT1[:-2]) # Numerical trick: Replace uz2[1] with an extrapolated value

    # Original correct one:
    # uz2[1:-1] = - r0 * cosl[1:-1]**2 * exp(-rkappa*(kmax-2.)/7.) * dDelta_PT1 / (cor1[1:-1,-2]**2 * aa * hh * dmdz)
    uz2[:] = - r0 * cosl[:]**2 * exp(-rkappa*(kmax-2.)/7.) * dDelta_PT1 / (cor1[:,-2]**2 * aa * hh * dmdz)

    #  **** Initialize the coefficients a,b,c,d,e,f ****
    c_a[:,:] = 1.0
    c_b[:,:] = 1.0
    c_c[:,1:-1] = dmdz**2 *ephalf[:,1:-1]*exp(-dz/(2*hh))  # This one should be correct
    c_d[:,1:-1] = dmdz**2 *ephalf[:,0:-2]*exp(dz/(2*hh)) # Check convention of ephalf
    c_e[:,1:-1] = -(c_a[:,1:-1]+c_b[:,1:-1]+c_c[:,1:-1]+c_d[:,1:-1])

    b = np.zeros((jmax1*kmax))

    row_index=[]
    col_index=[]
    coeff = []

    jrange = range(jmax1)
    krange = range(1,kmax-1)
    for j, k in itertools.product(jrange, krange):
    # for j in range(jmax1):
    #     for k in range(1,kmax-1):
        ind = input_jk_output_index(j,k,kmax)
        b[ind] = c_f[j,k]
        if (j<jmax1-1):
            # A[ind,input_jk_output_index(j+1,k,kmax)] = c_a[j,k]
            row_index.append(ind)
            col_index.append(input_jk_output_index(j+1,k,kmax))
            coeff.append(c_a[j,k])
        if (j>0):
            # A[ind,input_jk_output_index(j-1,k,kmax)] = c_b[j,k]
            row_index.append(ind)
            col_index.append(input_jk_output_index(j-1,k,kmax))
            coeff.append(c_b[j,k])
        # A[ind,input_jk_output_index(j,k+1,kmax)] = c_c[j,k]
        row_index.append(ind)
        col_index.append(input_jk_output_index(j,k+1,kmax))
        coeff.append(c_c[j,k])
        # A[ind,input_jk_output_index(j,k-1,kmax)] = c_d[j,k]
        row_index.append(ind)
        col_index.append(input_jk_output_index(j,k-1,kmax))
        coeff.append(c_d[j,k])
        # A[ind,input_jk_output_index(j,k,kmax)] = c_e[j,k]
        row_index.append(ind)
        col_index.append(input_jk_output_index(j,k,kmax))
        coeff.append(c_e[j,k])

    # ==== Upper boundary condition - thermal wind ====
    # for j in range(1,jmax1-1):
    for j in range(jmax1):
        ind1 = input_jk_output_index(j,kmax-1,kmax)
        b[ind1] = uz2[j] #- r0 * cosl[j]**2  * exp(-rkappa*(kmax-2.)/7.) * (Delta_PT1[j+1]-Delta_PT1[j-1])/ (cor1[j,-2]**2 * aa * hh * dmdz)
        # A[ind1,ind1] = 1.0
        row_index.append(ind1)
        col_index.append(ind1)
        coeff.append(1.0)
        # A[ind1,input_jk_output_index(j,kmax-3,kmax)] = -1.0
        row_index.append(ind1)
        col_index.append(input_jk_output_index(j,kmax-3,kmax))
        coeff.append(-1.0)

    # Try sparse matrix
    # print 'try sparse matrix'
    # A = csc_matrix((coeff_noslip, (row_index, col_index)), shape=(jmax1*kmax,jmax1*kmax))
    # print 'shape of A=',A.shape
    # print 'Does it work?'
#
#     csc_matrix((data, (row_ind, col_ind)), [shape=(M, N)])
#     where data, row_ind and col_ind satisfy the relationship a[row_ind[k], col_ind[k]] = data[k].


        # A[ind1,input_jk_output_index(j,kmax-3,kmax)] = -1.0
        #uz2[1:-1] = - r0 * cosl[1:-1]**2 * exp(-rkappa*(kmax-2.)/7.) * (Delta_PT1[2:]-Delta_PT1[:-2]) / (cor1[1:-1,-2]**2 * aa * hh * dmdz)

    # === Make a copy to deal with adiabatic boundary condition ===
    # A: no-slip
    # A_adiab: adiabatic boundary conditions
    row_index_adiab = copy(row_index)
    col_index_adiab = copy(col_index)
    coeff_adiab = copy(coeff)
    b_adiab = np.copy(b)

    # print 'does it work till here?'

    # A_adiab = np.copy(A)

    # ==== Lower boundary condition - adiabatic (k=0) ====
    for j in range(jmax1):
        ind0 = input_jk_output_index(j,0,kmax)
        b_adiab[ind0] = uz1[j]
        # A_adiab[ind0,ind0] = -1.0 # k=0
        row_index_adiab.append(ind0)
        col_index_adiab.append(ind0)
        coeff_adiab.append(-1.0)
        # A_adiab[ind0,input_jk_output_index(j,2,kmax)] = 1.0 # k=2
        row_index_adiab.append(ind0)
        col_index_adiab.append(input_jk_output_index(j,2,kmax))
        coeff_adiab.append(1.0)

    A_adiab = csc_matrix((coeff_adiab, (row_index_adiab, col_index_adiab)), shape=(jmax1*kmax,jmax1*kmax))

    # ==== Lower boundary condition - no-slip (k=0) ====
    for j in range(jmax1):
        ind = input_jk_output_index(j,0,kmax)
        b[ind] = zmu1[j,0]*cosl[j]/cor1[j,0]
        # A[ind,ind] = 1.0
        row_index.append(ind)
        col_index.append(ind)
        coeff.append(1.0)

    A = csc_matrix((coeff, (row_index, col_index)), shape=(jmax1*kmax,jmax1*kmax))

    # print 'is it ok till here????'

    # === Solving the linear system ===
    u2_adiab =  spsolve(A_adiab, b_adiab)
    u2 = spsolve(A, b)

    # === Mapping back to 2D matrix ===
    u_adiab = np.zeros((jmax1+2,kmax))
    u = np.zeros((jmax1+2,kmax))
    for j in range(jmax1):
        for k in range(kmax):
            u_adiab[j+1,k] = u2_adiab[j*kmax + k]
            u[j+1,k] = u2[j*kmax + k]

    u_MassCorr_adiab = np.zeros_like(u_adiab)
    u_MassCorr_noslip = np.zeros_like(u)
    # u_MassCorr[1:-1,:] = u[1:-1,:] * cor1[1:-1,:] / cosl[1:-1,np.newaxis]
    u_MassCorr_adiab[1:-1,:] = u_adiab[1:-1,:] * cor1 / cosl[:,np.newaxis]
    u_MassCorr_noslip[1:-1,:] = u[1:-1,:] * cor1 / cosl[:,np.newaxis]

    # --- Initialize T_MassCorr to be output ---
    u_Ref_regular_adiab = np.zeros_like(zmum)
    u_Ref_regular_noslip = np.zeros_like(zmum)
    u_MassCorr_regular_adiab = np.zeros_like(zmum)
    u_MassCorr_regular_noslip = np.zeros_like(zmum)
    T_Ref_regular_adiab = np.zeros_like(zmum)
    T_Ref_regular_noslip = np.zeros_like(zmum)
    T_MassCorr_regular_adiab = np.zeros_like(zmum)
    T_MassCorr_regular_noslip = np.zeros_like(zmum)

    for u_MassCorr,u_MassCorr_regular,u_Ref_regular,T_MassCorr_regular,T_Ref_regular,BCstring in \
    zip([u_MassCorr_adiab,u_MassCorr_noslip],\
        [u_MassCorr_regular_adiab,u_MassCorr_regular_noslip],\
        [u_Ref_regular_adiab,u_Ref_regular_noslip],\
        [T_MassCorr_regular_adiab,T_MassCorr_regular_noslip],\
        [T_Ref_regular_adiab,T_Ref_regular_noslip],\
        ['Adiabatic','Noslip']):

        # ---- Back out temperature correction here -----
        T_MassCorr = np.zeros_like(u_MassCorr)
        for k in range(1,kmax-2):
            for j in range(2,jmax1,2): # This is temperature not potential temperature!!! Need to check.
                # print 'alat['+str(j)+']=',alat[j]
                # T_MassCorr[j,k] = T_MassCorr[j-2,k] - (2.*om*gl[j])*aa*hh*dmdz / (r0 * cosl[j]) * (u_MassCorr[j,k+1]-u_MassCorr[j,k-1])
                T_MassCorr[j,k] = T_MassCorr[j-2,k] - (2.*om*gl[j-1])*aa*hh*dmdz / (r0 * cosl[j-1]) * (u_MassCorr[j-1,k+1]-u_MassCorr[j-1,k-1])
            # ---- First do interpolation (gl is regular grid) ----
            # f_Todd = interpolate.interp1d(gl[:-1:2],T_MassCorr[1:-1:2,k])    #[jmax x kmax]
            #f_Todd = interpolate.interp1d(gl_2[::2],T_MassCorr[::2,k])    #[jmax x kmax]
            #f_Todd_ex = extrap1d(f_Todd)

            f_Todd = interpolate.interp1d(gl_2[::2],T_MassCorr[::2,k],
                                          kind='linear',fill_value='extrapolate')
            T_MassCorr[:,k] = f_Todd(gl_2[:])



            # T_MassCorr[:,k] = f_Todd_ex(gl_2[:]) # Get all the points interpolated

            # ---- Then do domain average ----
            T_MC_mean = np.mean(T_MassCorr[:,k])
            T_MassCorr[:,k] -= T_MC_mean

        # --- First, interpolate MassCorr back to regular grid first ---
        f_u_MassCorr = interpolate.interp1d(alat_2,u_MassCorr,axis=0, kind='linear')    #[jmax x kmax]
        u_MassCorr_regular[:,-nlat//2:] = f_u_MassCorr(ylat[-nlat//2:]).T
        f_T_MassCorr = interpolate.interp1d(alat_2,T_MassCorr,axis=0, kind='linear')    #[jmax x kmax]
        T_MassCorr_regular[:,-nlat//2:] = f_T_MassCorr(ylat[-nlat//2:]).T

        u_Ref = zmum[:,-nlat//2:] - u_MassCorr_regular[:,-nlat//2:]
        T_ref = zm_PT[:,-nlat//2:] * np.exp(-np.arange(kmax)/7. * rkappa)[:,np.newaxis] - T_MassCorr_regular[:,-nlat//2:]

        u_Ref_regular[:,-nlat//2:] = u_Ref
        T_Ref_regular[:,-nlat//2:] = T_ref
#
           #plot_all_ref_quan = False
        if plot_all_ref_quan:
            # --- height coordinate ---
            height = np.array([i for i in range(kmax)]) # in [km]
            # --- Colorbar scale ---
            contour_int = np.arange(-120,145,5)
            dT_contour_int = np.arange(-120,81,5)
            T_contour_int = np.arange(160,321,5)
            # --- Start plotting figure ---
            fig = plt.subplots(figsize=(12,12))
            plt.subplot(221)
            plt.contourf(ylat[-nlat//2:],height[:-2],u_MassCorr_regular[:-2,-nlat//2:],contour_int)
            plt.colorbar()
            c1=plt.contour(ylat[-nlat//2:],height[:-2],u_MassCorr_regular[:-2,-nlat//2:],contour_int[::2],colors='k')
            plt.clabel(c1,c1.levels,inline=True, fmt='%d', fontsize=10)
            plt.title('$\Delta$ u '+tstamp)
            plt.ylabel('height (km)')
            plt.subplot(222)
            plt.contourf(ylat[-nlat//2:],height[:-2],u_Ref[:-2,:],contour_int)
            plt.colorbar()
            c2=plt.contour(ylat[-nlat//2:],height[:-2],u_Ref[:-2,:],contour_int[::2],colors='k')
            plt.clabel(c2,c2.levels,inline=True, fmt='%d', fontsize=10)
            plt.title('$u_{REF}$ ('+BCstring+' BC)')
            plt.subplot(223)
            plt.contourf(ylat[-nlat//2:],height[:-2],T_MassCorr_regular[:-2,-nlat//2:],dT_contour_int)
            plt.colorbar()
            c3=plt.contour(ylat[-nlat//2:],height[:-2],T_MassCorr_regular[:-2,-nlat//2:],dT_contour_int,colors='k')
            plt.clabel(c3,c3.levels,inline=True, fmt='%d', fontsize=10)
            plt.title('$\Delta$ T')
            plt.ylabel('height (km)')
            plt.subplot(224)
            plt.contourf(ylat[-nlat//2:],height[:-2],T_ref[:-2,:],T_contour_int)
            plt.colorbar()
            c4=plt.contour(ylat[-nlat//2:],height[:-2],T_ref[:-2,:],T_contour_int[::2],colors='k')
            plt.clabel(c4,c4.levels,inline=True, fmt='%d', fontsize=10)
            plt.title('$T_{REF}$')
            plt.ylabel('height (km)')
            plt.tight_layout()
            plt.show()
            #plt.savefig('/home/csyhuang/Dropbox/Research-code/Sep12_test3_'+BCstring+'_'+tstamp+'.png')
            plt.close()

    # This is for only outputing Delta_u and Uref for no-slip and adiabatic boundary conditions.
    return u_MassCorr_regular_noslip,u_Ref_regular_noslip,T_MassCorr_regular_noslip,T_Ref_regular_noslip, u_MassCorr_regular_adiab,u_Ref_regular_adiab,T_MassCorr_regular_adiab,T_Ref_regular_adiab

# --- As a test whether the function Solve_Uref is working ---
if __name__ == "__main__":

    import matplotlib.pyplot as plt
    import numpy as np

    nlat = 121
    kmax = 49
    jmax1 = nlat

    # The codes below is just for testing purpose
    tstamp = 'random'
    ylat = np.linspace(-90,90,121,endpoint=True)
    t1 = np.random.rand(nlat,kmax)+np.ones((nlat,kmax))*0.001
    t2 = np.random.rand(nlat,kmax)+np.ones((nlat,kmax))*0.001
    t3 = np.random.rand(nlat,kmax)+np.ones((nlat,kmax))*0.001
    Delta_PT = np.random.rand(nlat)+np.ones((nlat))*0.001
    zm_PT = np.random.rand(nlat,kmax)+np.ones((nlat,kmax))*0.001
    Input_B0 = np.random.rand(nlat)+np.ones((nlat))*0.001
    Input_B1 = np.random.rand(nlat)+np.ones((nlat))*0.001
    eh = np.random.rand(jmax1, kmax)+np.ones((jmax1, kmax))*0.001
    Delta_PT = np.sort(np.random.rand(jmax1))

    xxx = solve_uref_both_bc(tstamp,t1,t2,ylat,t3,Delta_PT,zm_PT,Input_B0,Input_B1,use_real_Data=True)
    print(xxx)



import os
import pytest

import numpy as np
from math import pi
from scipy.interpolate import interp1d

from falwa.constant import *
from falwa.oopinterface import QGFieldNH18

# === Parameters specific for testing the qgfield class ===
nlat = 121
nlon = 240
xlon = np.linspace(0, 2. * pi, nlon, endpoint=False)
ylat = np.linspace(-90., 90., nlat, endpoint=True)
plev = np.array([1000,  900,  800,  700,  600,  500,  400,  300,  200, 100,   10,    1])
nlev = plev.size
kmax = 49
get_cwd = os.path.dirname(os.path.abspath(__file__))
u_field = np.reshape(np.loadtxt(get_cwd + '/data/global_demo_u.txt'), [nlev, nlat, nlon])
v_field = np.reshape(np.loadtxt(get_cwd + '/data/global_demo_v.txt'), [nlev, nlat, nlon])
t_field = np.reshape(np.loadtxt(get_cwd + '/data/global_demo_t.txt'), [nlev, nlat, nlon])
theta_field = t_field * (plev[:, np.newaxis, np.newaxis] / P_GROUND) ** (-DRY_GAS_CONSTANT / CP)


def test_qgfield():

    # Create a QGFieldNH18 object out of some masked array for testing. The test below is to ensure
    # warning is raised for masked array.
    with pytest.warns(UserWarning):
        qgfield = QGFieldNH18(
            xlon=xlon, ylat=np.ma.masked_equal(ylat, 0.0), plev=plev,
            u_field=np.ma.masked_equal(u_field, 0.0), v_field=np.ma.masked_equal(v_field, 0.0),
            t_field=np.ma.masked_equal(t_field, 0.0), kmax=kmax, maxit=100000, dz=1000., npart=None,
            tol=1.e-5, rjac=0.95, scale_height=SCALE_HEIGHT, cp=CP, dry_gas_constant=DRY_GAS_CONSTANT,
            omega=EARTH_OMEGA, planet_radius=EARTH_RADIUS, northern_hemisphere_results_only=True)

    # Check that the input fields are interpolated onto a grid of correct dimension
    # and the interpolated values are bounded.
    qgpv, interpolated_u, interpolated_v, interpolated_theta, static_stability = \
        qgfield.interpolate_fields()

    # Check that the dimensions of the interpolated fields are correct
    assert (kmax, nlat, nlon) == qgpv.shape
    assert (kmax, nlat, nlon) == interpolated_u.shape
    assert (kmax, nlat, nlon) == interpolated_v.shape
    assert (kmax, nlat, nlon) == interpolated_theta.shape
    assert (kmax,) == static_stability.shape

    assert (interpolated_u[1:-1, :, :].max() <= u_field.max()) & \
           (interpolated_u[1:-1, :, :].max() >= u_field.min())
    assert (interpolated_u[1:-1, :, :].min() <= u_field.max()) & \
           (interpolated_u[1:-1, :, :].min() >= u_field.min())
    assert (interpolated_v[1:-1, :, :].max() <= v_field.max()) & \
           (interpolated_v[1:-1, :, :].max() >= v_field.min())
    assert (interpolated_v[1:-1, :, :].min() <= v_field.max()) & \
           (interpolated_v[1:-1, :, :].min() >= v_field.min())
    assert (interpolated_theta[1:-1, :, :].max() <= theta_field.max()) & \
           (interpolated_theta[1:-1, :, :].max() >= theta_field.min())
    assert (interpolated_theta[1:-1, :, :].min() <= theta_field.max()) & \
           (interpolated_theta[1:-1, :, :].min() >= theta_field.min())
    assert 0 == np.isnan(qgpv).sum()
    assert 0 == (qgpv == float('Inf')).sum()

    # Check that the output reference states are of correct dimension, and
    # the QGPV reference state is non-decreasing.
    qref_north_hem, uref_north_hem, ptref_north_hem = \
        qgfield.compute_reference_states()

    # Check dimension of the input field
    assert (kmax, nlat // 2 + 1) == qref_north_hem.shape
    assert (kmax, nlat // 2 + 1) == uref_north_hem.shape
    assert (kmax, nlat // 2 + 1) == ptref_north_hem.shape

    # Check if qref is monotonically increasing in N. Hem
    assert (np.diff(qref_north_hem, axis=-1)[1:-1, 1:-1] >= 0.).all()

    # Check numerical values at maybe 45N
    np.testing.assert_allclose(qref_north_hem[1:-1, 30], np.array([
       1.38131594e-04, 1.19859978e-04, 1.05491689e-04, 9.93276096e-05,
       9.22312472e-05, 1.15428025e-04, 1.76390117e-04, 2.59869051e-04,
       2.68715711e-04, 2.38162745e-04, 2.18186744e-04, 1.86275673e-04,
       1.68824676e-04, 1.56962190e-04, 9.88995968e-05, 7.78025576e-05,
       7.11415702e-05, 8.75867385e-05, 8.90332624e-05, 8.90447767e-05,
       8.88902142e-05, 8.85380061e-05, 8.81466043e-05, 8.77005638e-05,
       8.73810977e-05, 8.70058817e-05, 8.67982961e-05, 8.65848763e-05,
       8.64754241e-05, 8.63178982e-05, 9.19433192e-05, 9.59023756e-05,
       9.09137157e-05, 8.63107880e-05, 8.63246689e-05, 8.76919204e-05,
       8.96391234e-05, 9.18857274e-05, 9.40389724e-05, 9.59951480e-05,
       9.83253872e-05, 1.00259563e-04, 1.02466469e-04, 1.04852489e-04,
       1.06957177e-04, 1.09106934e-04, 1.10848619e-04]), rtol=1.e-4)
    np.testing.assert_allclose(uref_north_hem[1:-1, 30], np.array([
       4.0792756,  7.9502687, 11.6558895, 15.18368, 18.527643,
       21.687384, 24.602285, 27.075981, 28.873444, 29.869736,
       30.134935, 30.241474, 30.34288, 30.507023, 30.719505,
       30.744917, 30.567951, 30.155119, 29.801447, 29.569159,
       29.49587, 29.563114, 29.779781, 30.163935, 30.708176,
       31.40156, 32.236263, 33.202354, 34.27791, 35.433662,
       36.633213, 38.06651, 39.856266, 41.948883, 44.131187,
       46.426163, 48.819324, 51.259815, 53.726505, 56.18757,
       58.665966, 61.160248, 63.64206, 66.105194, 68.56923,
       71.03153, 73.52619]), rtol=1.e-4)
    np.testing.assert_allclose(ptref_north_hem[1:-1, 30], np.array([
        282.9861, 289.11166, 294.43488, 298.87775, 302.74335,
        305.4141, 307.44733, 310.67563, 318.58322, 329.07108,
        339.64813, 352.70694, 366.25687, 379.23846, 389.83984,
        408.4669, 442.00266, 466.65228, 492.23962, 518.2976,
        544.6688, 571.2527, 597.9376, 624.6638, 651.401,
        678.0977, 704.8375, 731.67053, 758.4729, 785.2347,
        811.9185, 836.7763, 887.0521, 945.8203, 1005.51154,
        1065.4585, 1125.3812, 1185.4506, 1245.7306, 1306.2148,
        1366.9077, 1427.694, 1488.4418, 1549.4957, 1610.9202,
        1672.3165, 1733.3434]), rtol=1.e-4)

    # Check LWA and fluxes
    lwa_and_fluxes = qgfield.compute_lwa_and_barotropic_fluxes()
    assert lwa_and_fluxes.adv_flux_f1.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.adv_flux_f2.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.adv_flux_f3.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.convergence_zonal_advective_flux.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.divergence_eddy_momentum_flux.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.meridional_heat_flux.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.lwa_baro.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.u_baro.shape == (nlat // 2 + 1, nlon)
    assert lwa_and_fluxes.lwa.shape == (kmax, nlat // 2 + 1, nlon)

    # Check values at midlatitudes, maybe an average of 30-60 deg
    np.testing.assert_allclose(
        lwa_and_fluxes.lwa_baro[20:40, :].mean(axis=-1), np.array([
            19.630522, 21.613638, 23.790615, 25.932142, 28.172373, 30.730598, 33.59619, 36.23143, 38.318214, 40.161243,
            42.13259, 44.062523, 45.303726, 45.888157, 46.33659, 46.821716, 46.8923, 46.248104, 45.019547, 43.42369]),
        rtol=1.e-4)
    np.testing.assert_allclose(
        lwa_and_fluxes.meridional_heat_flux[20:40, :].mean(axis=-1), np.array([
            1.1194464e-05, 1.6376511e-05, 2.3858540e-05, 3.2586056e-05, 3.9103314e-05, 4.2148498e-05, 4.5597390e-05,
            5.1362793e-05, 5.7720707e-05, 6.1747254e-05, 6.7472894e-05, 7.2880815e-05, 6.9412905e-05, 6.0302500e-05,
            5.6556517e-05, 5.2262596e-05, 4.7150170e-05, 4.1664603e-05, 3.7316819e-05, 3.1581640e-05]), atol=1.e-8)
    np.testing.assert_allclose(
            lwa_and_fluxes.divergence_eddy_momentum_flux[20:40, :].mean(axis=-1), np.array([
            -2.2376915e-05, -2.8482620e-05, -2.9509038e-05, -2.8688662e-05,
            -3.0097528e-05, -2.2539192e-05, -1.0400371e-05, -4.7039898e-06,
            -4.3675382e-06, -7.4572531e-06, -7.7944615e-06, -5.3719723e-06,
            -1.5880020e-07,  8.9033156e-06,  2.3703995e-05,  3.7202812e-05,
            3.2651566e-05,  1.3700411e-05, -4.6122295e-08,  4.2839401e-06]), atol=1.e-8)


def test_qgfield_full_globe():

    # Create a QGFieldNH18 object for testing
    qgfield = QGFieldNH18(
        xlon=xlon, ylat=ylat, plev=plev, u_field=u_field, v_field=v_field, t_field=t_field, kmax=kmax,
        maxit=100000, dz=1000., npart=None, tol=1.e-5, rjac=0.95, scale_height=SCALE_HEIGHT, cp=CP,
        dry_gas_constant=DRY_GAS_CONSTANT, omega=EARTH_OMEGA, planet_radius=EARTH_RADIUS,
        northern_hemisphere_results_only=False)

    # Check that the input fields are interpolated onto a grid of correct dimension
    # and the interpolated values are bounded.
    interpolated_fields = qgfield.interpolate_fields()

    # Check that the output reference states are of correct dimension, and
    # the QGPV reference state is non-decreasing.
    qref_full_hem, uref_full_hem, ptref_full_hem = qgfield.compute_reference_states()

    # Check dimension of the input field
    assert (kmax, nlat) == qref_full_hem.shape
    assert (kmax, nlat) == uref_full_hem.shape
    assert (kmax, nlat) == ptref_full_hem.shape

    # Check if qref is monotonically increasing in both hemisphere (exclude equator)
    assert (np.diff(qref_full_hem, axis=-1)[1:-1, 1:nlat//2-1] >= 0.).all()  # South Hem
    assert (np.diff(qref_full_hem, axis=-1)[1:-1, nlat // 2 + 1:-1] >= 0.).all()  # North Hem

    # Check LWA and fluxes
    lwa_and_fluxes = qgfield.compute_lwa_and_barotropic_fluxes()
    assert lwa_and_fluxes.adv_flux_f1.shape == (nlat, nlon)
    assert lwa_and_fluxes.adv_flux_f2.shape == (nlat, nlon)
    assert lwa_and_fluxes.adv_flux_f3.shape == (nlat, nlon)
    assert lwa_and_fluxes.convergence_zonal_advective_flux.shape == (nlat, nlon)
    assert lwa_and_fluxes.divergence_eddy_momentum_flux.shape == (nlat, nlon)
    assert lwa_and_fluxes.meridional_heat_flux.shape == (nlat, nlon)
    assert lwa_and_fluxes.lwa_baro.shape == (nlat, nlon)
    assert lwa_and_fluxes.u_baro.shape == (nlat, nlon)
    assert lwa_and_fluxes.lwa.shape == (kmax, nlat, nlon)


def test_raise_error_for_unrealistic_fields():
    """
    Error shall be raised if the SOR algorithm for computing reference state does not converge.
    """
    qgfield = QGFieldNH18(
        xlon=xlon, ylat=ylat, plev=plev, u_field=u_field, v_field=u_field, t_field=u_field, kmax=kmax,
        maxit=100000, dz=1000., npart=None, tol=1.e-5, rjac=0.95, scale_height=SCALE_HEIGHT, cp=CP,
        dry_gas_constant=DRY_GAS_CONSTANT, omega=EARTH_OMEGA, planet_radius=EARTH_RADIUS)
    qgfield.interpolate_fields()
    with pytest.raises(ValueError):
        qgfield.compute_reference_states()


def test_raise_error_for_unrealistic_kmax():
    """
    Error shall be raised if kmax is out of bound of the pressure grid provided in the input data

    .. versionadded:: 0.4.0
    """
    too_large_kmax = 1000
    with pytest.raises(ValueError):
        QGFieldNH18(xlon=xlon, ylat=ylat, plev=plev, u_field=u_field, v_field=v_field, t_field=t_field, kmax=too_large_kmax)


def test_interpolate_fields_even_grids():
    """
    To test whether the new features of even-to-odd grid interpolation works well.

    .. versionadded:: 0.3.5

    """
    ylat_even = np.linspace(-90., 90., nlat + 1, endpoint=True)[1:-1]
    u_field_even = interp1d(ylat, u_field, axis=1,
                            fill_value="extrapolate")(ylat_even)
    v_field_even = interp1d(ylat, v_field, axis=1,
                            fill_value="extrapolate")(ylat_even)
    t_field_even = interp1d(ylat, t_field, axis=1,
                            fill_value="extrapolate")(ylat_even)

    # Create a QGFieldNH18 object for testing
    qgfield_even = QGFieldNH18(
        xlon=xlon, ylat=ylat_even, plev=plev, u_field=u_field_even, v_field=v_field_even,
        t_field=t_field_even, kmax=kmax, maxit=100000, dz=1000., npart=None, tol=1.e-5, rjac=0.95,
        scale_height=SCALE_HEIGHT, cp=CP, dry_gas_constant=DRY_GAS_CONSTANT, omega=EARTH_OMEGA,
        planet_radius=EARTH_RADIUS)

    qgpv, interpolated_u, interpolated_v, interpolated_theta, static_stability = \
        qgfield_even.interpolate_fields()

    assert kmax == qgfield_even.kmax
    assert nlat-1 == qgfield_even.get_latitude_dim()
    assert nlon == qgfield_even.nlon

    # Check that the dimensions of the interpolated fields are correct
    assert (kmax, nlat-1, nlon) == qgpv.shape
    assert (kmax, nlat-1, nlon) == interpolated_u.shape
    assert (kmax, nlat-1, nlon) == interpolated_v.shape
    assert (kmax, nlat-1, nlon) == interpolated_theta.shape
    assert (kmax,) == static_stability.shape

    # Check that at the interior grid points, the interpolated fields
    # are bounded
    assert (interpolated_u[1:-1, 1:-1, 1:-1].max() <= u_field.max()) & \
           (interpolated_u[1:-1, 1:-1, 1:-1].max() >= u_field.min())

    assert (interpolated_u[1:-1, 1:-1, 1:-1].min() <= u_field.max()) & \
           (interpolated_u[1:-1, 1:-1, 1:-1].min() >= u_field.min())

    assert (interpolated_v[1:-1, 1:-1, 1:-1].max() <= v_field.max()) & \
           (interpolated_v[1:-1, 1:-1, 1:-1].max() >= v_field.min())

    assert (interpolated_v[1:-1, 1:-1, 1:-1].min() <= v_field.max()) & \
           (interpolated_v[1:-1, 1:-1, 1:-1].min() >= v_field.min())

    assert (interpolated_theta[1:-1, 1:-1, 1:-1].max() <= theta_field.max()) & \
           (interpolated_theta[1:-1, 1:-1, 1:-1].max() >= theta_field.min())

    assert (interpolated_theta[1:-1, 1:-1, 1:-1].min() <= theta_field.max()) & \
           (interpolated_theta[1:-1, 1:-1, 1:-1].min() >= theta_field.min())

    assert 0 == np.isnan(qgpv).sum()
    assert 0 == (qgpv == float('Inf')).sum()




"""
------------------------------------------
File name: wrapper.py
Author: Clare Huang
"""
import numpy as np
from falwa.constant import EARTH_RADIUS
from falwa import basis


def barotropic_eqlat_lwa(ylat, vort, area, dmu, n_points, planet_radius=EARTH_RADIUS):
    """
    Compute equivalent latitude and wave activity on a barotropic sphere.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    qref : numpy.array
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values;
                    dimension = [nlat_s x nlon]

    Examples
    --------
    :doc:`Example Notebook <notebooks/Example_barotropic>`
    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]
    if n_points is None:
        n_points = nlat

    qref, fawa = basis.eqvlat_fawa(
        ylat, vort, area, n_points)
    lwa_result, _ = basis.lwa(nlon, nlat, vort, qref, dmu)

    return qref, lwa_result


def barotropic_input_qref_to_compute_lwa(ylat, qref, vort, area, dmu, planet_radius=EARTH_RADIUS):

    """
    This function computes LWA based on a *prescribed* Qref instead of Qref obtained from the vorticity field on a barotropic sphere.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    qref : numpy.array
        1-d numpy array of prescribed reference value of vorticity at each latitude; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values; dimension = [nlat_s x nlon]
    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]
    lwa_result = basis.lwa(nlon, nlat, vort, qref, dmu)
    return lwa_result


def eqvlat_hemispheric(ylat, vort, area, nlat_s=None, n_points=None, planet_radius=EARTH_RADIUS):

    """
    Compute equivalent latitude in a hemispheric domain.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat_s*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    q_part : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).

    """
    nlat = vort.shape[0]
    qref = np.zeros(nlat)

    if nlat_s is None:
        nlat_s = nlat // 2

    if n_points is None:
        n_points = nlat_s

    # --- Southern Hemisphere ---
    qref1, _ = basis.eqvlat_fawa(ylat[:nlat_s], vort[:nlat_s, :], area[:nlat_s, :],
                            n_points, planet_radius=planet_radius)
    qref[:nlat_s] = qref1

    # --- Northern Hemisphere ---
    vort2 = -vort[::-1, :]
    qref2, _ = basis.eqvlat_fawa(ylat[:nlat_s], vort2[:nlat_s, :], area[:nlat_s, :],
                            n_points, planet_radius=planet_radius)
    qref[-nlat_s:] = qref2[::-1]

    return qref


def eqvlat_bracket_hemispheric(ylat, vort, area, nlat_s=None, n_points=None, planet_radius=EARTH_RADIUS, vgrad=None):

    """
    Compute equivalent latitude and <...>_Q in Nakamura and Zhu (2010) in a hemispheric domain.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat_s*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.
    vgrad: numpy.ndarray, optional
        2-d numpy array of laplacian (or higher-order laplacian) values; dimension = (nlat, nlon)

    Returns
    -------
    q_part : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    brac : numpy.ndarray or None
        1-d numpy array of averaged vgrad in the square bracket.
        If *vgrad* = None, *brac* = None.

    """
    nlat = vort.shape[0]
    qref = np.zeros(nlat)
    brac = np.zeros(nlat)

    if nlat_s is None:
        nlat_s = nlat // 2

    if n_points is None:
        n_points = nlat_s

    # --- Southern Hemisphere ---
    qref1, brac1 = basis.eqvlat_fawa(ylat[:nlat_s], vort[:nlat_s, :],
                                area[:nlat_s, :],
                                n_points, planet_radius=planet_radius,
                                vgrad=vgrad)
    qref[:nlat_s] = qref1

    # --- Northern Hemisphere ---
    vort2 = -vort[::-1, :]
    qref2, brac2 = basis.eqvlat_fawa(ylat[:nlat_s], vort2[:nlat_s, :],
                                area[:nlat_s, :],
                                n_points, planet_radius=planet_radius,
                                vgrad=vgrad)

    qref[-nlat_s:] = qref2[::-1]

    brac[:nlat_s] = brac1
    brac[-nlat_s:] = brac2[::-1]

    return qref, brac


def qgpv_eqlat_lwa(ylat, vort, area, dmu, nlat_s=None, n_points=None, planet_radius=EARTH_RADIUS):

    """
    Compute equivalent latitutde *qref* and local wave activity *lwa_result* based
    on Quasi-geostrophic potential vorticity field *vort* at a pressure level as
    outlined in Huang and Nakamura (2017).

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of Quasi-geostrophic potential vorticity field; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat_s*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    qref : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values;
                    dimension = [nlat_s x nlon]

    Examples
    --------
    :doc:`Example Notebook <notebooks/Example_qgpv>`
    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]

    if nlat_s is None:
        nlat_s = nlat // 2

    if n_points is None:
        n_points = nlat_s

    qref = np.zeros(nlat)
    lwa_result = np.zeros((nlat, nlon))

    # --- Southern Hemisphere ---
    qref1, _ = basis.eqvlat_fawa(
        ylat[:nlat_s], vort[:nlat_s, :], area[:nlat_s, :], n_points, planet_radius=planet_radius)
    qref[:nlat_s] = qref1
    lwa_result[:nlat_s, :], _ = basis.lwa(nlon, nlat_s,
                                          vort[:nlat_s, :],
                                          qref1, dmu[:nlat_s])

    # --- Northern Hemisphere ---
    vort2 = -vort[::-1, :]
    # Added the minus sign, but gotta see if NL_North is affected
    qref2, _ = basis.eqvlat_fawa(
        ylat[:nlat_s], vort2[:nlat_s, :], area[:nlat_s, :], n_points, planet_radius=planet_radius)
    qref[-nlat_s:] = -qref2[::-1]
    lwa_result[-nlat_s:, :], _ = basis.lwa(nlon, nlat_s,
                                           vort[-nlat_s:, :],
                                           qref[-nlat_s:],
                                           dmu[-nlat_s:])
    return qref, lwa_result


def qgpv_eqlat_lwa_ncforce(ylat, vort, ncforce, area, dmu, nlat_s=None, n_points=None, planet_radius=EARTH_RADIUS):

    """
    Compute equivalent latitutde *qref*, local wave activity *lwa_result* and
    non-conservative force on wave activity *capsigma* based on Quasi-
    geostrophic potential vorticity field *vort* at a pressure level as
    outlined in Huang and Nakamura (2017).

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of Quasi-geostrophic potential vorticity field; dimension = (nlat, nlon).
    ncforce: numpy.ndarray
        2-d numpy array of non-conservative force field (i.e. theta in NZ10(a) in equation (23a) and (23b));
        dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat_s*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    qref : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values; dimension = (nlat, nlon).
    capsigma: numpy.ndarray
        2-d numpy array of non-conservative force contribution value; dimension = (nlat, nlon).

    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]

    if nlat_s is None:
        nlat_s = nlat // 2

    if n_points is None:
        n_points = nlat_s

    qref = np.zeros(nlat)
    lwa_result = np.zeros((nlat, nlon))
    capsigma = np.zeros((nlat, nlon))

    # --- Southern Hemisphere ---
    qref1, _ = basis.eqvlat_fawa(ylat[:nlat_s],
                            vort[:nlat_s, :], area[:nlat_s, :],
                            n_points, planet_radius=planet_radius)
    qref[:nlat_s] = qref1
    lwa_result[:nlat_s, :], \
    capsigma[:nlat_s, :] = basis.lwa(nlon, nlat_s, vort[:nlat_s, :],
                                     qref1, dmu[:nlat_s],
                                     ncforce=ncforce[:nlat_s, :])

    # --- Northern Hemisphere ---
    vort2 = -vort[::-1, :]
    # Added the minus sign, but gotta see if NL_North is affected
    qref2, _ = basis.eqvlat_fawa(ylat[:nlat_s], vort2[:nlat_s, :], area[:nlat_s, :],
                            n_points, planet_radius=planet_radius)
    qref[-nlat_s:] = -qref2[::-1]
    lwa_result[-nlat_s:, :], \
    capsigma[-nlat_s:, :] = basis.lwa(nlon, nlat_s, vort[-nlat_s:, :],
                                      qref[-nlat_s:], dmu[-nlat_s:],
                                      ncforce=ncforce[-nlat_s:, :])
    return qref, lwa_result, capsigma


def qgpv_eqlat_lwa_options(ylat, vort, area, dmu, nlat_s=None, n_points=None, vgrad=None, ncforce=None,
                           planet_radius=EARTH_RADIUS):

    """
    Compute equivalent latitutde *qref*, local wave activity *lwa_result* and
    non-conservative force on wave activity *capsigma* based on Quasi-
    geostrophic potential vorticity field *vort* at a pressure level as
    outlined in Huang and Nakamura (2017).

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : numpy.ndarray
        2-d numpy array of Quasi-geostrophic potential vorticity field; dimension = (nlat, nlon).
    ncforce: numpy.ndarray
        2-d numpy array of non-conservative force field (i.e. theta in NZ10(a) in equation (23a) and (23b));
        dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    n_points : int, default None
        Analysis resolution to calculate equivalent latitude. If input as None, it will be initialized as *nlat_s*.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    return_dict : dictionary
        A dictionary that consist of the 4 computed outputs listed below.
    (1) qref : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    (2) brac_result: numpy.ndarray
        1-d numpy array of <...>_Q(y) in NZ10 where latitude y is given by ylat; dimension = (nlat).
    (3) lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values; dimension = (nlat, nlon).
    (4) capsigma: numpy.ndarray
        2-d numpy array of non-conservative force contribution value; dimension = (nlat, nlon).

    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]

    if nlat_s is None:
        nlat_s = nlat // 2

    if n_points is None:
        n_points = nlat_s

    qref = np.zeros(nlat)
    lwa_result = np.zeros((nlat, nlon))
    if ncforce is not None:
        capsigma = np.zeros((nlat, nlon))
    if vgrad is not None:
        brac_result = np.zeros(nlat)

    # --- Southern Hemisphere ---
    if vgrad is None:
        qref1, brac = basis.eqvlat_fawa(ylat[:nlat_s], vort[:nlat_s, :],
                                   area[:nlat_s, :],
                                   n_points, planet_radius=planet_radius)
    else:
        qref1, brac = basis.eqvlat_fawa(ylat[:nlat_s], vort[:nlat_s, :],
                                   area[:nlat_s, :],
                                   n_points, planet_radius=planet_radius,
                                   vgrad=vgrad[:nlat_s, :])

    qref[:nlat_s] = qref1
    if vgrad is not None:
        brac_result[:nlat_s] = brac

    if ncforce is not None:
        lwa_result[:nlat_s, :], capsigma[:nlat_s, :] = \
        basis.lwa(nlon, nlat_s,
                  vort[:nlat_s, :],
                  qref1, dmu[:nlat_s],
                  ncforce=ncforce[:nlat_s, :])
    else:
        lwa_result[:nlat_s, :], _ = basis.lwa(nlon, nlat_s, vort[:nlat_s, :],
                                              qref1, dmu[:nlat_s])

    # --- Northern Hemisphere ---
    vort2 = -vort[::-1, :]
    # Added the minus sign, but gotta see if NL_North is affected

    if vgrad is None:
        qref2, brac = basis.eqvlat_fawa(ylat[:nlat_s], vort2[:nlat_s, :],
                                   area[:nlat_s, :],
                                   n_points, planet_radius=planet_radius)
    else:
        vgrad2 = -vgrad[::-1, :]  # Is this correct?
        qref2, brac = basis.eqvlat_fawa(ylat[:nlat_s], vort2[:nlat_s, :],
                                   area[:nlat_s, :],
                                   n_points, planet_radius=planet_radius,
                                   vgrad=vgrad2[:nlat_s, :])

    qref[-nlat_s:] = -qref2[::-1]
    if vgrad is not None:
        brac_result[-nlat_s:] = -brac[::-1]

    if ncforce is not None:
        lwa_result[-nlat_s:, :], \
        capsigma[-nlat_s:, :] = basis.lwa(nlon, nlat_s, vort[-nlat_s:, :],
                                          qref[-nlat_s:],
                                          dmu[-nlat_s:],
                                          ncforce=ncforce[-nlat_s:, :])
    else:
        lwa_result[-nlat_s:, :], _ = basis.lwa(nlon, nlat_s, vort[-nlat_s:, :],
                                               qref[-nlat_s:], dmu[-nlat_s:])

    # Return things as dictionary
    return_dict = dict()
    return_dict['qref'] = qref
    return_dict['lwa_result'] = lwa_result

    if vgrad is not None:
        return_dict['brac_result'] = brac_result
    if ncforce is not None:
        return_dict['capsigma'] = capsigma

    return return_dict


def qgpv_input_qref_to_compute_lwa(ylat, qref, vort, area, dmu, nlat_s=None, planet_radius=EARTH_RADIUS):
    """
    Compute equivalent latitutde *qref* and local wave activity *lwa_result* based
    on Quasi-geostrophic potential vorticity field *vort* at a pressure level as
    outlined in Huang and Nakamura (2017). This function computes lwa based on a
    prescribed *qref* instead of *qref* obtained from the QGPV field.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    qref : numpy.ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    vort : numpy.ndarray
        2-d numpy array of Quasi-geostrophic potential vorticity field; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    lwa_result : numpy.ndarray
        2-d numpy array of local wave activity values; dimension = (nlat, nlon).
    """
    nlat = vort.shape[0]
    nlon = vort.shape[1]
    if nlat_s is None:
        nlat_s = nlat // 2

    lwa_result = np.zeros((nlat, nlon))

    # --- Southern Hemisphere ---
    lwa_result[:nlat_s, :], dummy = basis.lwa(nlon, nlat_s, vort[:nlat_s, :],
                                              qref[:nlat_s], dmu[:nlat_s])

    # --- Northern Hemisphere ---
    lwa_result[-nlat_s:, :], dummy = basis.lwa(nlon, nlat_s, vort[-nlat_s:, :],
                                               qref[-nlat_s:], dmu[-nlat_s:])

    return lwa_result


def theta_lwa(ylat, theta, area, dmu, nlat_s=None, n_points=None, planet_radius=EARTH_RADIUS):
    """
    Compute the surface wave activity *B* based on surface potential temperature.
    See Nakamura and Solomon (2010a) for details.

    Parameters
    ----------
    ylat : numpy.array
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    theta : numpy.ndarray
        2-d numpy array of surface potential temperature field; dimension = (nlat, nlon).
    area : numpy.ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    dmu: numpy.array
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    nlat_s : int, default None
        The index of grid point that defines the extent of hemispheric domain from the pole. If input as None, it will be initialize as nlat // 2.
    planet_radius : float, default 6.378e+6
        radius of spherical planet of interest consistent with input 'area'.

    Returns
    -------
    qref : numpy.array
        1-d numpy array of value reference potential temperature *Theta(y)* approximated by box counting method, where latitude y is given by ylat; dimension = (nlat).
    lwa_result : numpy.ndarray
        2-d numpy array of local surface wave activity values; dimension = (nlat, nlon).

    """
    nlat = theta.shape[0]
    nlon = theta.shape[1]
    if nlat_s is None:
        nlat_s = nlat // 2
    if n_points is None:
        n_points = nlat_s

    qref = np.zeros(nlat)
    lwa_result = np.zeros((nlat, nlon))

    # --- southern Hemisphere ---
    qref[:nlat_s], brac = basis.eqvlat_fawa(ylat[:nlat_s], theta[:nlat_s, :],
                                       area[:nlat_s, :], n_points,
                                       planet_radius=planet_radius)
    lwa_result[:nlat_s, :], dummy = basis.lwa(nlon, nlat_s, theta[:nlat_s, :],
                                              qref[:nlat_s], dmu[:nlat_s])

    # --- northern Hemisphere ---
    theta2 = theta[::-1, :]
    # Added the minus sign, but gotta see if NL_north is affected
    qref2, brac = basis.eqvlat_fawa(ylat[:nlat_s], theta2[:nlat_s, :],
                               area[:nlat_s, :],
                               n_points, planet_radius=planet_radius)
    qref[-nlat_s:] = qref2[::-1]
    lwa_result[-nlat_s:, :], dummy = basis.lwa(nlon, nlat_s,
                                               theta[-nlat_s:, :],
                                               qref[-nlat_s:],
                                               dmu[-nlat_s:])

    return qref, lwa_result



"""
------------------------------------------
File name: xarrayinterface.py
Author: Christopher Polster
"""
import functools
import numpy as np
import xarray as xr

from falwa import __version__
from falwa.oopinterface import QGFieldNH18


def _is_ascending(arr):
    return np.all(np.diff(arr) > 0)

def _is_descending(arr):
    return np.all(np.diff(arr) < 0)

def _is_equator(x):
    return abs(x) < 1.0e-4

# Coordinate name lookup
_NAMES_PLEV = ["plev", "lev", "level", "isobaricInhPa"]
_NAMES_YLAT = ["ylat", "lat", "latitude"]
_NAMES_XLON = ["xlon", "lon", "longitude"]
_NAMES_TIME = ["time", "date", "datetime"]
# Wind and temperature name lookup
_NAMES_U = ["u", "U"]
_NAMES_V = ["v", "V"]
_NAMES_T = ["t", "T"]
# Budget terms name lookup
_NAMES_LWA  = ["lwa_baro"]
_NAMES_CZAF = ["convergence_zonal_advective_flux"]
_NAMES_DEMF = ["divergence_eddy_momentum_flux"]
_NAMES_MHF  = ["meridional_heat_flux"]


def _get_dataarray(data, names, user_names=None):
    name = _get_name(data, names, user_names=user_names)
    return data[name]


def _get_name(ds, names, user_names=None):
    # If the first name from the list of defaults is in the user-provided
    # dictionary, use the name provided there
    if user_names is not None and names[0] in user_names:
        name = user_names[names[0]]
        if name not in ds:
            raise KeyError(f"specified variable '{name}' not found")
        return name
    # Else, search in default list of names
    for name in names:
        if name in ds:
            return name
    raise KeyError(f"no matching variable for '{names[0]}' found")


class _MetadataServiceProvider:
    """Metadata services for the QGDataset

    The class provides metadata from its own registry and can be instanciated
    to provide additional metadata based on a template QGField object and
    user-provided information about additional non-core dimensions.

    Parameters
    ----------
    field : QGField
        Template QGField object to extract metadata from.
    other_coords : None | dict
        Mapping of dimension name to dimension coordinates of non-core
        dimensions. Entries must reflect order of dimensions.
    """

    def __init__(self, field, other_coords=None):
        self.field = field
        # Depend on dict to preserve ordering of dims (Python 3.7+)
        self.other_coords = dict(other_coords) if other_coords is not None else dict()

    @property
    def other_dims(self):
        """Names of non-core dimensions"""
        return tuple(self.other_coords.keys())

    @property
    def other_shape(self):
        """Shape of non-core dimensions"""
        return tuple(value.size for value in self.other_coords.values())

    @property
    def other_size(self):
        """Size of non-core dimensions"""
        return np.product(self.other_shape)

    # numpy convenience functions

    def shape(self, var):
        """Shape of a variable (non-core and core dims)"""
        shape = list(self.other_shape)
        # Get sizes of field dimensions from template field
        for name in self.info(var)["dim_names"]:
            shape.append(getattr(self.field, name).size)
        return tuple(shape)

    def flatten_other(self, arr):
        """Flatten the non-core dimensions of the array"""
        n = len(self.other_shape)
        assert arr.shape[:n] == self.other_shape, f"expected other shape of {self.other_shape}"
        return arr.reshape((self.other_size, *arr.shape[n:]))

    def restore_other(self, arr):
        """Un-flatten the non-core dimensions of the array"""
        assert arr.shape[0] == self.other_size, f"expected other size of {self.other_size}"
        return arr.reshape(self.other_shape + arr.shape[1:])

    # xarray convenience functions

    def dims(self, var):
        """Dimension names (non-core and core dims)"""
        return self.other_dims + self.info(var)["core_dims"]

    def coords(self, var):
        """Coordinate dictionary (non-core and core dims)"""
        coords = self.other_coords.copy()
        info = self.info(var)
        for dim, name in zip(info["core_dims"], info["dim_names"]):
            coords[dim] = getattr(self.field, name)
        return coords

    def as_dataarray(self, arr, var):
        """Create a DataArray from the input array as the given variable"""
        arr = np.asarray(arr)
        if arr.shape != self.shape(var):
            arr = self.restore_other(arr)
        assert arr.shape == self.shape(var)
        return xr.DataArray(
            arr,
            dims=self.dims(var),
            coords=self.coords(var),
            name=var,
            attrs=self.attrs(var)
        )

    def attrs(self, var=None):
        """Attributes for a Dataset (var=None) or a DataArray (var!=None)"""
        if var is not None:
            return self.info(var)["attrs"]
        return {
            "kmax": self.field.kmax,
            "dz": self.field.dz,
            "maxit": self.field.maxit,
            "tol": self.field.tol,
            "npart": self.field.npart,
            "rjac": self.field.rjac,
            "scale_height": self.field.scale_height,
            "cp": self.field.cp,
            "dry_gas_constant": self.field.dry_gas_constant,
            "omega": self.field.omega,
            "planet_radius": self.field.planet_radius,
            "prefactor": self.field.prefactor,
            "protocol": type(self.field).__name__,
            "package": f"hn2016_falwa {__version__}"
        }

    # General information from a variable registry
    # (must be kept up-to-date with oopinterface.QGField, see below)

    _VARS = dict()

    @classmethod
    def register_var(cls, var, core_dims, dim_names=None, attrs=None):
        """Add a new variable configuration to the registry

        Parameters
        ----------
        var : string
            Name of the variable in the registry.
        core_dims : Tuple[string]
            Core dimensions of the variable, i.e. the fundamental dimensions
            that a single field of this variable always has. Core dimensions
            must always be the last dimensions in the array.
        dim_names : Tuple[string], optional
            Name overrides for data access on the QGField template object.
        attrs : dict, optional
            Attributes for the variable, attached to any produced DataArray.
        """
        cls._VARS[var] = {
            "core_dims": core_dims,
            "dim_names": dim_names if dim_names is not None else core_dims,
            "attrs": attrs
        }

    @classmethod
    def info(cls, var):
        """Metadata information from the variable registry"""
        return cls._VARS[var]


# Interpolated fields
_MetadataServiceProvider.register_var("qgpv", ("height", "ylat", "xlon"))
_MetadataServiceProvider.register_var("interpolated_u", ("height", "ylat", "xlon"))
_MetadataServiceProvider.register_var("interpolated_v", ("height", "ylat", "xlon"))
_MetadataServiceProvider.register_var("interpolated_theta", ("height", "ylat", "xlon"))
_MetadataServiceProvider.register_var("static_stability", ("height",))
_MetadataServiceProvider.register_var("static_stability_n", ("height",))
_MetadataServiceProvider.register_var("static_stability_s", ("height",))
# Reference state fields (y-z cross section)
_MetadataServiceProvider.register_var("qref", ("height", "ylat"), dim_names=("height","ylat_ref_states"))
_MetadataServiceProvider.register_var("uref", ("height", "ylat"), dim_names=("height", "ylat_ref_states"))
_MetadataServiceProvider.register_var("ptref", ("height", "ylat"), dim_names=("height", "ylat_ref_states"))
# Column-averaged fields (x-y horizontal fields)
_MetadataServiceProvider.register_var("u_baro", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("lwa_baro", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("adv_flux_f1", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("adv_flux_f2", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("adv_flux_f3", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("convergence_zonal_advective_flux", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("divergence_eddy_momentum_flux", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
_MetadataServiceProvider.register_var("meridional_heat_flux", ("ylat", "xlon"), dim_names=("ylat_ref_states", "xlon"))
# 3-dimensional LWA (full x-y-z fields)
_MetadataServiceProvider.register_var("lwa", ("height", "ylat", "xlon"), dim_names=("height", "ylat_ref_states", "xlon"))


class _DataArrayCollector(property):
    # Getter properties for DataArray-based access to QGField properties.
    # Inherits from property, so instances are recognized as properties by
    # sphinx for the docs.

    def __init__(self, var):
        self.var = var
        self.__doc__ = (
            f"See :py:attr:`oopinterface.QGFieldBase.{self.var}`."
            "\n\nReturns\n-------\nxarray.DataArray"
        )

    def __get__(self, qgds, objtype=None):
        arr = np.asarray([getattr(field, self.var) for field in qgds.fields])
        return qgds.metadata.as_dataarray(arr, self.var)


class QGDataset:
    """A wrapper for multiple QGField objects with xarray in- and output.

    For each combination of timestep, ensemble member, etc. in the input data,
    a :py:class:`oopinterface.QGField` object is instanciated. The constructor
    will automatically flip latitude and pressure dimensions of the input data
    if necessary to meet the requirements of QGField.

    This wrapper class imitates the methods of QGField (but not the
    properties/attributes) and collects and re-organizes output data in xarray
    Datasets for convenience. All calculations are performed by the QGField
    routines.

    .. versionadded:: 0.6.1

    Parameters
    ----------
    da_u : xarray.DataArray | xarray.Dataset
        Input 3D fields of zonal wind. The 3D fields's dimensions must end with
        height, latitude and longitude. Other dimensions (e.g. time, ensemble
        member id) are preserved in the output datasets.
        Alternatively, a dataset can be given, from which `u`, `v` and `T`
        fields are then extracted. The `da_v` and `da_t` arguments can then be
        omitted or used as an override.
    da_v : xarray.DataArray, optional
        Input 3D fields of meridional wind. The 3D fields's dimensions must end
        with height, latitude and longitude. Other dimensions (e.g. time,
        ensemble member id) are preserved in the output datasets.
    da_t : xarray.DataArray, optional
        Input 3D fields of temperature. The 3D fields's dimensions must end
        with height, latitude and longitude. Other dimensions (e.g. time,
        ensemble member id) are preserved in the output datasets.
    var_names : dict, optional
        If the auto-detection of variable or coordinate names fails, provide
        a lookup table that maps `plev`, `ylat`, `xlon`, `u`, `v` and/or `t` to
        the names used in the dataset.
    qgfield : QGField class, optional
        The QGField class to use in the computation. Default:
        :py:class:`oopinterface.QGFieldNH18`.
    qgfield_args : tuple, optional
        Positional arguments given to the QGField constructor.
    qgfield_kwargs : dict, optional
        Keyword arguments given to the QGField constructor.

    Examples
    -------
    >>> data = xarray.open_dataset("path/to/some/uvt-data.nc")
    >>> qgds = QGDataset(data)
    >>> qgds.interpolate_fields()
    <xarray.Dataset> ...

    :doc:`notebooks/demo_script_for_nh2018_with_xarray`

    >>> data_u = xarray.load_dataset("path/to/some/u-data.nc")
    >>> data_v = xarray.load_dataset("path/to/some/v-data.nc")
    >>> data_t = xarray.load_dataset("path/to/some/t-data.nc")
    >>> qgds = QGDataset(data_u, data_v, data_t)
    """

    def __init__(self, da_u, da_v=None, da_t=None, *, var_names=None,
                 qgfield=QGFieldNH18, qgfield_args=None, qgfield_kwargs=None):
        if var_names is None:
            var_names = dict()

        # Check input data type first
        assert isinstance(da_u, xr.Dataset) or isinstance(da_u, xr.DataArray)
        assert da_v is None or isinstance(da_v, xr.DataArray) or isinstance(da_v, xr.Dataset)
        assert da_t is None or isinstance(da_t, xr.DataArray) or isinstance(da_t, xr.Dataset)

        # Also support construction from single-arg and mixed variants
        if isinstance(da_u, xr.Dataset):
            # Fill up missing DataArrays for v and t from the Dataset but give
            # priority to existing v and t fields from the args
            if da_v is None:
                da_v = _get_dataarray(da_u, _NAMES_V, var_names)
            elif isinstance(da_v, xr.Dataset):
                da_v = _get_dataarray(da_v, _NAMES_V, var_names)
            # else: assume da_v is dataarray so there is no issue

            if da_t is None:
                da_t = _get_dataarray(da_u, _NAMES_T, var_names)
            elif isinstance(da_t, xr.Dataset):
                da_t = _get_dataarray(da_t, _NAMES_T, var_names)
            # else: assume da_t is dataarray so there is no issue

            # Always take u
            da_u = _get_dataarray(da_u, _NAMES_U, var_names)

        # Assertions about da_u, da_v, da_t
        assert da_u is not None, "missing u field"
        assert da_v is not None, "missing v field"
        assert da_t is not None, "missing t field"
        # Assign standard names to the input fields
        da_u = da_u.rename("u")
        da_v = da_v.rename("v")
        da_t = da_t.rename("t")
        # Merge into one dataset and keep the reference. xarray will avoid
        # copying the data in the merge, so the operation should be relatively
        # cheap and fast. The merge further verifies that the coordinates of
        # the three DataArrays match.
        self._ds = xr.merge([da_u, da_v, da_t], join="exact", compat="equals")
        # QGField* configuration
        self._qgfield = qgfield
        self._qgfield_args = list() if qgfield_args is None else qgfield_args
        self._qgfield_kwargs = dict() if qgfield_kwargs is None else qgfield_kwargs
        # Extract spatial coordinates
        da_plev = _get_dataarray(self._ds.coords, _NAMES_PLEV, var_names)
        da_ylat = _get_dataarray(self._ds.coords, _NAMES_YLAT, var_names)
        da_xlon = _get_dataarray(self._ds.coords, _NAMES_XLON, var_names)
        # Check that field coordinates end in lev, lat, lon
        assert da_u.dims[-3] == da_plev.name, f"dimension -3 of input fields must be '{da_plev.name}' (plev)"
        assert da_u.dims[-2] == da_ylat.name, f"dimension -2 of input fields must be '{da_ylat.name}' (ylat)"
        assert da_u.dims[-1] == da_xlon.name, f"dimension -1 of input fields must be '{da_xlon.name}' (xlon)"
        assert da_u.dims == da_v.dims, f"dimensions of fields '{da_u.name}' (u) and '{da_v.name}' (v) don't match"
        assert da_u.dims == da_t.dims, f"dimensions of fields '{da_u.name}' (u) and '{da_t.name}' (t) don't match"
        # The input data may contain multiple time steps, ensemble members etc.
        # Flatten all these other dimensions so a single loop covers all
        # fields. These dimensions are restored in the output datasets.
        other_dims = da_u.dims[:-3]
        other_shape = tuple(da_u[dim].size for dim in other_dims)
        other_size = np.product(other_shape, dtype=np.int64)
        _shape = (other_size, *da_u.shape[-3:])
        # Extract value arrays and collapse all additional dimensions
        u = da_u.data.reshape(_shape)
        v = da_v.data.reshape(_shape)
        t = da_t.data.reshape(_shape)
        # Automatically determine how fields need to be flipped so they match
        # the requirements of QGField and extract coordinate values
        flip = []
        # Ensure that ylat is ascending
        ylat = da_ylat.values
        if not _is_ascending(ylat):
            ylat = np.flip(ylat)
            flip.append(-2)
        # Ensure that plev is descending
        plev = da_plev.values
        if not _is_descending(plev):
            plev = np.flip(plev)
            flip.append(-3)
        # Ordering of xlon doesn't matter here
        xlon = da_xlon.values
        # Create a QGField object for each combination of timestep, ensemble
        # member, etc.
        self._fields = []
        for u_field, v_field, t_field in zip(u, v, t):
            # Apply reordering to fields
            if flip:
                u_field = np.flip(u_field, axis=flip)
                v_field = np.flip(v_field, axis=flip)
                t_field = np.flip(t_field, axis=flip)
            field = self._qgfield(xlon, ylat, plev, u_field, v_field, t_field,
                                  *self._qgfield_args, **self._qgfield_kwargs)
            self._fields.append(field)
        # Make sure there is at least one field in the dataset
        assert self._fields, "empty input"
        # Tailored metadata access
        self.metadata = _MetadataServiceProvider(self._fields[0], other_coords={
            dim: self._ds.coords[dim] for dim in other_dims
        })

    @property
    def fields(self):
        """Access to the QGField objects created by the QGDataset.

        The :py:class:`.oopinterface.QGField` objects are stored in a flattened
        list.
        """
        return self._fields

    @property
    def attrs(self):
        """Metadata dictionary that is attached to output datasets."""
        return self.metadata.attrs()

    def interpolate_fields(self, return_dataset=True):
        """Call `interpolate_fields` on all contained fields.

        See :py:meth:`.oopinterface.QGFieldBase.interpolate_fields`.

        .. note::
            A QGField class may define static stability globally or
            hemispherically on each level. The output dataset contains a single
            variable for static stability in case of a global definition and
            two variables for static stability for a hemispheric definition
            (suffix ``_n`` for the northern hemisphere and ``_s`` for the
            southern hemisphere).

        Parameters
        ----------
        return_dataset : bool
            Whether to return the computed fields as a dataset.

        Returns
        -------
        xarray.Dataset or None
        """
        for field in self._fields:
            field.interpolate_fields(return_named_tuple=False)
        if return_dataset:
            data_vars = {
                "qgpv": self.qgpv,
                "interpolated_u": self.interpolated_u,
                "interpolated_v": self.interpolated_v,
                "interpolated_theta": self.interpolated_theta
            }
            # Stability property may contain multiple variables
            stability = self.static_stability
            if isinstance(stability, xr.DataArray):
                stability = (stability,)
            data_vars.update({ s.name: s for s in stability })
            return xr.Dataset(data_vars, attrs=self.attrs)

    # Accessors for individual field properties computed in interpolate_fields
    qgpv = _DataArrayCollector("qgpv")
    interpolated_u = _DataArrayCollector("interpolated_u")
    interpolated_v = _DataArrayCollector("interpolated_v")
    interpolated_theta = _DataArrayCollector("interpolated_theta")

    @property
    def static_stability(self):
        """See :py:attr:`oopinterface.QGFieldBase.static_stability`.

        Returns
        -------
        xr.Dataset | Tuple[xr.Dataset, xr.Dataset]
        """
        stability = np.asarray([getattr(field, "static_stability") for field in self._fields])
        if stability.ndim == 2:
            # One vertical profile of static stability per field: global
            return self.metadata.as_dataarray(stability, "static_stability")
        elif stability.ndim == 3 and stability.shape[-2] == 2:
            # Two vertical profiles of static stability per field: hemispheric
            return (
                self.metadata.as_dataarray(stability[:,0,:], "static_stability_n"),
                self.metadata.as_dataarray(stability[:,1,:], "static_stability_s")
            )
        else:
            raise ValueError(f"cannot process shape of returned static stability field: {stability.shape}")


    def compute_reference_states(self, return_dataset=True):
        """Call `compute_reference_states` on all contained fields.

        See :py:meth:`.oopinterface.QGFieldBase.compute_reference_states`.

        Parameters
        ----------
        return_dataset : bool
            Whether to return the computed fields as a dataset.

        Returns
        -------
        xarray.Dataset or None
        """
        for field in self._fields:
            field.compute_reference_states(return_named_tuple=False)
        if return_dataset:
            data_vars = {
                "qref": self.qref,
                "uref": self.uref,
                "ptref": self.ptref,
            }
            return xr.Dataset(data_vars, attrs=self.attrs)

    # Accessors for individual field properties computed in compute_reference_states
    qref = _DataArrayCollector("qref")
    uref = _DataArrayCollector("uref")
    ptref = _DataArrayCollector("ptref")

    def compute_lwa_and_barotropic_fluxes(self, return_dataset=True):
        """Call `compute_lwa_and_barotropic_fluxes` on all contained fields.

        See :py:meth:`.oopinterface.QGFieldBase.compute_lwa_and_barotropic_fluxes`.

        Parameters
        ----------
        return_dataset : bool
            Whether to return the computed fields as a dataset.

        Returns
        -------
        xarray.Dataset or None
        """
        for field in self._fields:
            field.compute_lwa_and_barotropic_fluxes(return_named_tuple=False)
        if return_dataset:
            data_vars = {
                "adv_flux_f1": self.adv_flux_f1,
                "adv_flux_f2": self.adv_flux_f2,
                "adv_flux_f3": self.adv_flux_f3,
                "convergence_zonal_advective_flux": self.convergence_zonal_advective_flux,
                "divergence_eddy_momentum_flux": self.divergence_eddy_momentum_flux,
                "meridional_heat_flux": self.meridional_heat_flux,
                "lwa_baro": self.lwa_baro,
                "u_baro": self.u_baro,
                "lwa": self.lwa,
            }
            return xr.Dataset(data_vars, attrs=self.attrs)

    # Accessors for individual field properties computed in compute_lwa_and_barotropic_fluxes
    adv_flux_f1 = _DataArrayCollector("adv_flux_f1")
    adv_flux_f2 = _DataArrayCollector("adv_flux_f2")
    adv_flux_f3 = _DataArrayCollector("adv_flux_f3")
    convergence_zonal_advective_flux = _DataArrayCollector("convergence_zonal_advective_flux")
    divergence_eddy_momentum_flux = _DataArrayCollector("divergence_eddy_momentum_flux")
    meridional_heat_flux = _DataArrayCollector("meridional_heat_flux")
    lwa_baro = _DataArrayCollector("lwa_baro")
    u_baro = _DataArrayCollector("u_baro")
    lwa = _DataArrayCollector("lwa")



def integrate_budget(ds, var_names=None):
    """Compute the integrated LWA budget terms for the given data.

    Integrates the LWA tendencies from equation (2) of `NH18
    <https://doi.org/10.1126/science.aat0721>`_ in time (over the time interval
    covered by the input data). The residual (term IV) is determined by
    subtracting terms (I), (II) and (III) from the LWA difference between the
    last and first time step in the data. Uses
    :py:meth:`xarray.DataArray.integrate` for the time integration of the
    tendencies.

    See :py:meth:`QGDataset.compute_lwa_and_barotropic_fluxes`, which computes
    all required tendency terms as well as the LWA fields.

    .. versionadded:: 0.6.1

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset containing the budget tendencies for the time integration
        interval.
    var_names : dict, optional
        The names of LWA and the tendency term variables are automatically
        detected. If the auto-detection fails, provide a lookup table that maps
        `time`, `lwa_baro`, `convergence_zonal_advective_flux`,
        `divergence_eddy_momentum_flux`, and/or `meridional_heat_flux` to the
        names used in the input dataset.

    Returns
    -------
    xarray.Dataset

    Examples
    -------
    >>> qgds = QGDataset(data)
    >>> ...
    >>> terms = qgds.compute_lwa_and_barotropic_fluxes()
    >>> integrate_budget(terms.isel({ "time": slice(5, 10) }))
    """
    name_time = _get_name(ds, _NAMES_TIME, var_names)
    name_lwa  = _get_name(ds, _NAMES_LWA,  var_names)
    name_czaf = _get_name(ds, _NAMES_CZAF, var_names)
    name_demf = _get_name(ds, _NAMES_DEMF, var_names)
    name_mhf  = _get_name(ds, _NAMES_MHF,  var_names)
    # Integration time interval covered by the data
    start = ds[name_time].values[0]
    stop = ds[name_time].values[-1]
    # Determine the change in LWA over the time interval
    dlwa = ds[name_lwa].sel({ name_time: stop }) - ds[name_lwa].sel({ name_time: start })
    # Integrate the known tendencies in time
    czaf = ds[name_czaf].integrate(coord=name_time, datetime_unit="s")
    demf = ds[name_demf].integrate(coord=name_time, datetime_unit="s")
    mhf  = ds[name_mhf].integrate(coord=name_time, datetime_unit="s")
    # Compute the residual from the difference between the explicitly computed
    # budget terms and the actual change in LWA
    res  = dlwa - czaf - demf - mhf
    # Include all 5 integrated budget terms in the output
    data_vars = {
        "delta_lwa": dlwa,
        "integrated_convergence_zonal_advective_flux": czaf,
        "integrated_divergence_eddy_momentum_flux": demf,
        "integrated_meridional_heat_flux": mhf,
        "residual": res
    }
    # Copy attributes from original dataset and add information about
    # integration interval (start and end timestamps as well as integration
    # time interval in seconds)
    attrs = dict(ds.attrs)
    attrs["integration_start"] = str(start)
    attrs["integration_stop"] = str(stop)
    attrs["integration_seconds"] = (stop - start) / np.timedelta64(1000000000)
    return xr.Dataset(data_vars, ds.coords, attrs)


def hemisphere_to_globe(ds, var_names=None):
    """Create a global dataset from a hemispheric one.

    Takes data from the given hemisphere, mirrors it to the other hemisphere
    and combines both hemispheres into a global dataset.

    If the meridional wind component is found in the dataset, its values will
    be negated on the created hemisphere. This results in identical fields of
    local wave activity on both hemispheres (since absolute vorticity is also
    the same except for the sign), making it possible to use
    `northern_hemisphere_only` in the methods of :py:class:`QGDataset` even if
    only southern hemisphere data is available. Discontinuities in the
    meridional wind and derived fields arise due to this at the equator but
    they generally have only a small effect on the outputs.

    .. versionadded:: 0.6.1

    Parameters
    ----------
    ds : xarray.Dataset
        Input dataset. Must contain the equator (0° latitude).
    var_names : dict, optional
        The names of the latitude and meridional wind fields are automatically
        detected. If the auto-detection of the latitude coordinate and/or the
        meridional wind component fails, provide a lookup table that maps
        `ylat`, and/or `v` to the names used in the dataset.

    Returns
    -------
    xarray.Dataset
    """
    # Determine if the northern or southern hemisphere is present
    ylat_name = _get_name(ds, _NAMES_YLAT, var_names)
    eq0 = _is_equator(ds[ylat_name][0])
    assert eq0 or _is_equator(ds[ylat_name][-1]), (
        "equator not found on the hemisphere; "
        "make sure latitudes either begin or end with 0° latitude"
    )
    # Flip the data along ylat and omit the equator which should not appear
    # twice in the output
    flipped_noeq = slice(None, 0, -1) if eq0 else slice(-2, None, -1)
    sd = ds.reindex({ ylat_name: ds[ylat_name][flipped_noeq] })
    # Latitudes are now on the other hemisphere
    sd[ylat_name] = -sd[ylat_name]
    # Also flip the meridional wind (if present in the dataset). This results
    # in mirrored LWA fields on both hemispheres, the discontinuities this
    # creates on the equator are acceptable.
    try:
        v_name = _get_name(ds, _NAMES_V, var_names)
        sd[v_name] = -sd[v_name]
    except KeyError:
        pass
    # Assemble global dataset
    return xr.concat([sd, ds] if eq0 else [ds, sd], dim=ylat_name)



"""
------------------------------------------
File name: basis.py
Author: Clare Huang
"""
from typing import Tuple, Optional
import numpy as np
from math import pi


def eqvlat_fawa(
        ylat, vort, area, n_points, planet_radius=6.378e+6,
        output_fawa=True) -> Tuple[np.ndarray, Optional[np.ndarray]]:

    """
    Compute equivalent latitude and finite amplitude wave activity with the full

    Parameters
    ----------
    ylat : sequence or array_like
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    n_points: int
        Analysis resolution to calculate equivalent latitude.
    planet_radius: float
        Radius of spherical planet of interest consistent with input *area*. Default: earth's radius 6.378e+6
    output_fawa: bool
        Whether to output FAWA. If True, return tuple would consist of (qref, fawa). Else, this function will
        return (qref, None).

    Returns
    -------
    qref : ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    fawa : ndarray or None
        1-d finite-amplitude local wave activity
    """
    vort_min, vort_max = np.amin(vort), np.amax(vort)
    qbar = vort.mean(axis=-1)  # zonal mean vorticity
    vort_bins = np.linspace(vort_min, vort_max, n_points, endpoint=True)  # uniform vort bins
    vort_flat, area_flat = vort.flatten(), area.flatten()  # Flatten the 2D arrays to 1D

    # *** Find equivalent latitude ***
    inds = np.digitize(vort_flat, vort_bins)  # sort vorticity values into bins
    aq = np.cumsum([0] + [area_flat[np.where(inds == i)].sum() for i in range(1, vort_bins.size)])  # cumulative area
    cq = np.cumsum([0] + [np.multiply(area_flat, vort_flat)[np.where(inds == i)].sum()  # cumulative vort * area
                          for i in range(1, vort_bins.size)])
    eqv_gaussian_grid = aq / (2 * pi * planet_radius ** 2) - 1.0  # ascending from -1
    eqv_latitude = np.rad2deg(np.arcsin(eqv_gaussian_grid))
    qref = np.interp(ylat, eqv_latitude, vort_bins)  # ascending. qref = np.interp(alat[::-1], aq, q_part_u) also works

    # *** Compute FAWA ***
    if output_fawa:
        cref = np.interp(ylat, eqv_latitude, cq)
        cbar = np.zeros_like(cref)
        for j in np.arange(ylat.size-2, 0, -1):
            cbar[j] = cbar[j+1] + 0.5 * (qbar[j+1]*area.sum(axis=1)[j+1] + qbar[j]*area.sum(axis=1)[j])
        fawa = -(cbar[1:] + cref[1:])/(2 * pi * planet_radius)
        return qref, fawa

    return qref, None


def eqvlat_vgrad(ylat, vort, area, n_points, planet_radius=6.378e+6, vgrad=None):

    """
    Compute equivalent latitude, and optionally <...>_Q in Nakamura and Zhu (2010).

    Parameters
    ----------
    ylat : sequence or array_like
        1-d numpy array of latitude (in degree) with equal spacing in ascending order; dimension = nlat.
    vort : ndarray
        2-d numpy array of vorticity values; dimension = (nlat, nlon).
    area : ndarray
        2-d numpy array specifying differential areal element of each grid point; dimension = (nlat, nlon).
    n_points: int
        Analysis resolution to calculate equivalent latitude.
    planet_radius: float
        Radius of spherical planet of interest consistent with input *area*. Default: earth's radius 6.378e+6
    vgrad: ndarray, optional
        2-d numpy array of laplacian (or higher-order laplacian) values; dimension = (nlat, nlon)

    Returns
    -------
    q_part : ndarray
        1-d numpy array of value Q(y) where latitude y is given by ylat; dimension = (nlat).
    brac : ndarray or None
        1-d numpy array of averaged vgrad in the square bracket.
        If *vgrad* = None, *brac* = None.

    """

    vort_min = np.min([vort.min(), vort.min()])
    vort_max = np.max([vort.max(), vort.max()])
    q_part_u = np.linspace(vort_min, vort_max, n_points, endpoint=True)
    aa = np.zeros(q_part_u.size)  # to sum up area
    vort_flat = vort.flatten()  # Flatten the 2D arrays to 1D
    area_flat = area.flatten()

    if vgrad is not None:
        dp = np.zeros_like(aa)
        vgrad_flat = vgrad.flatten()

    # Find equivalent latitude:
    inds = np.digitize(vort_flat, q_part_u)
    for i in np.arange(0, aa.size):  # Sum up area in each bin
        aa[i] = np.sum(area_flat[np.where(inds == i)])
        if vgrad is not None:
            # This is to avoid the divided-by-zero error
            if aa[i] == 0.:
                continue
            else:
                dp[i] = np.sum(area_flat[np.where(inds == i)] *
                               vgrad_flat[np.where(inds == i)]) / aa[i]

    aq = np.cumsum(aa)
    if vgrad is not None:
        brac = np.zeros_like(aa)
        brac[1:-1] = 0.5 * (dp[:-2] + dp[2:])

    y_part = aq / (2 * pi * planet_radius**2) - 1.0
    lat_part = np.rad2deg(np.arcsin(y_part))
    q_part = np.interp(ylat, lat_part, q_part_u)

    if vgrad is not None:
        brac_return = np.interp(ylat, lat_part, brac)
    else:
        brac_return = None

    return q_part, brac_return


def lwa(nlon, nlat, vort, q_part, dmu, ncforce=None):

    """
    At each grid point of vorticity q(x,y) and reference state vorticity Q(y),
    this function calculate the difference between the line integral of [q(x,y+y')-Q(y)]
    (and ncforce, if given) over the domain {y+y'>y,q(x,y+y')<Q(y)} and {y+y'<y,q(x,y+y')>Q(y)}.
    See fig. (1) and equation (13) of Huang and Nakamura (2016).
    dmu is a vector of length nlat: dmu = cos(phi) d(phi) such that phi is the latitude.


    Parameters
    ----------
    nlon : int
        Longitudinal dimension of vort (i.e. vort.shape[1]).
    nlat : int
        Latitudinal dimension of vort (i.e. vort.shape[0]).
    vort : ndarray
        2-d numpy array of vorticity values; dimension = (nlat,nlon).
    Q_part: sequence or array_like
        1-d numpy array of Q (vorticity reference state) as a function of latitude. Size = nlat.
    dmu: sequence or array_like
        1-d numpy array of latitudinal differential length element (e.g. dmu = planet_radius * cos(lat) d(lat)). Size = nlat.
    ncforce: ndarray or None, optional
        2-d numpy array of non-conservative force field (i.e. theta in NZ10(a) in equation (23a) and (23b))


    Returns
    -------
    lwact : ndarray
        2-d numpy array of local wave activity calculated; dimension = (nlat,nlon).
    bigsigma : ndarray or None
        2-d numpy array of the nonconservative forces acting on local wave activity computed from *ncforce*.
        If *ncforce* = None, *bigsigma* = None.

    """

    lwact = np.zeros((nlat, nlon))
    if ncforce is not None:
        bigsigma = np.zeros((nlat, nlon))

    for j in np.arange(0, nlat - 1):
        vort_e = vort[:, :] - q_part[j]
        vort_boo = np.zeros((nlat, nlon))
        vort_boo[np.where(vort_e[:, :] < 0)] = -1
        vort_boo[:j + 1, :] = 0
        vort_boo[np.where(vort_e[:j + 1, :] > 0)] = 1
        lwact[j, :] = np.sum(vort_e * vort_boo *
                             dmu[:, np.newaxis], axis=0)
        if ncforce is not None:
            bigsigma[j, :] = np.sum(ncforce * vort_boo *
                                    dmu[:, np.newaxis], axis=0)

    if ncforce is None:
        bigsigma = None

    return lwact, bigsigma



"""
------------------------------------------
File name: data_storage.py
Author: Clare Huang
"""
from typing import Tuple, Optional, Union, List, NamedTuple
import numpy as np


class InvalidCallOfSHemVariables(Exception):
    """
    northern_hemisphere_results_only = True. 
    Southern hemispheric variables are not computed.
    """


class HemisphericProperty:

    def __init__(self, attr, ndims_fill=(0, 0), doc=None):
        self.attr = attr
        self.__doc__ = doc
        # Dimensions to fill in pre- and post-latitude
        npre, npost = ndims_fill
        self.slc_pre  = tuple(slice(None) for _ in range(npre))
        self.slc_post = tuple(slice(None) for _ in range(npost))

    @property
    def ndim_j(self):
        return len(self.slc_pre)


class NHemProperty(HemisphericProperty):

    def __get__(self, obj, objtype=None):
        if obj.northern_hemisphere_results_only:
            return getattr(obj, self.attr)
        # Assemble full slice from slices of pre-latitude dimensions, latitude
        # dimension slice and post-latitude dimensions slices
        slc = (*self.slc_pre, slice(-(obj.nlat//2+1), None), *self.slc_post)
        return getattr(obj, self.attr)[slc]

    def __set__(self, obj, value):
        jdim = value.shape[self.ndim_j]
        slc = (*self.slc_pre, slice(-jdim, None), *self.slc_post)
        getattr(obj, self.attr)[slc] = value


class SHemProperty(HemisphericProperty):

    def __get__(self, obj, objtype=None):
        if obj.northern_hemisphere_results_only:
            raise InvalidCallOfSHemVariables()
        slc = (*self.slc_pre, slice(None, obj.nlat//2+1), *self.slc_post)
        return getattr(obj, self.attr)[slc]

    def __set__(self, obj, value):
        jdim = value.shape[self.ndim_j]
        if obj.northern_hemisphere_results_only:
            raise InvalidCallOfSHemVariables()
        slc = (*self.slc_pre, slice(None, jdim), *self.slc_post)
        vlc = (*self.slc_pre, slice(None, None, -1), *self.slc_post)
        getattr(obj, self.attr)[slc] = value[vlc]



class DerivedQuantityStorage:
    """
    This class manages the storage of derived variables in :py:class:`oopinterface.QGField`.

    Variables are stored in fortran indexing order for easy communication with
    f2py modules. To return variables in python indexing order, use the method
    :py:meth:`fortran_to_python` to swap the axes.
    """
    def __init__(
        self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
        swapaxis_1: int, swapaxis_2: int, northern_hemisphere_results_only: bool):
        self._pydim = pydim  # python dimension
        self._fdim = fdim    # fortran dimension
        self._swapaxis_1 = swapaxis_1
        self._swapaxis_2 = swapaxis_2
        self._northern_hemisphere_results_only = northern_hemisphere_results_only

    @property
    def pydim(self):
        return self._pydim

    @property
    def fdim(self):
        return self._fdim

    @property
    def swapaxis_1(self):
        return self._swapaxis_1

    @property
    def swapaxis_2(self):
        return self._swapaxis_2

    @property
    def northern_hemisphere_results_only(self):
        return self._northern_hemisphere_results_only

    def fortran_to_python(self, phy_field: np.ndarray):
        return np.swapaxes(phy_field, self.swapaxis_1, self.swapaxis_2)

    def python_to_fortran(self, phy_field: np.ndarray):  # This may not be necessary
        return np.swapaxes(phy_field, self.swapaxis_2, self.swapaxis_1)


class DomainAverageStorage(DerivedQuantityStorage):
    """
    This stores global/hemispheric averaged potential temperature and static stability.

    Fortran dimension: (kmax)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int, northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        self.t0: Optional[np.array] = None
        self.tn0: Optional[np.array] = None
        self.ts0: Optional[np.array] = None
        self.static_stability: Optional[np.array] = None
        self.static_stability_n: Optional[np.array] = None
        self.static_stability_s: Optional[np.array] = None


class InterpolatedFieldsStorage(DerivedQuantityStorage):
    """
    This class stores 3D fields on interpolated grids, including:

    - u
    - v
    - theta (potential temperature)
    - avort (absolute vorticity, used as boundary condition to solve for reference state in NHN22)

    Fortran dimension: (nlon, nlat, kmax)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int, northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        self.interpolated_u: Optional[np.ndarray] = None
        self.interpolated_v: Optional[np.ndarray] = None
        self.interpolated_theta: Optional[np.ndarray] = None
        self.interpolated_avort: Optional[np.ndarray] = None
        self.qgpv: Optional[np.ndarray] = None


class ReferenceStatesStorage(DerivedQuantityStorage):
    """
    This class stores height-latitude 2D fields on interpolated grids, including:

    - uref
    - qref (it actually stores qref/f, where f is Coriolis paramter)
    - tref (potential temperature reference state)

    Fortran dimension: (nlat, kmax)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int,
                 northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        self.qref = np.zeros(self.fdim)  # This is to substitute self._qref_ntemp
        self.uref = np.zeros(self.fdim)
        self.ptref = np.zeros(self.fdim)
        kmax, self.nlat = self.pydim

    qref_nhem = NHemProperty("qref", (0, 1))
    qref_shem = SHemProperty("qref", (0, 1))

    def qref_correct_unit(self, ylat, omega, python_indexing=True):
        """
        This returns Qref of the correct unit to the user.
        TODO: encapsulate this elsewhere to avoid potential error
        """
        qref_right_unit = \
            self.qref * 2 * omega * np.sin(np.deg2rad(ylat[:, np.newaxis]))
        if python_indexing:
            return self.fortran_to_python(qref_right_unit) # (kmax, nlat)
        return qref_right_unit

    uref_nhem = NHemProperty("uref", (0, 1))
    uref_shem = SHemProperty("uref", (0, 1))

    ptref_nhem = NHemProperty("ptref", (0, 1))
    ptref_shem = SHemProperty("ptref", (0, 1))


class LWAStorage(DerivedQuantityStorage):
    """
    This class stores 3D LWA field on interpolated grids.

    Fortran dimension: (nlon, nlat, kmax)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int,
                 northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        self.lwa = np.zeros(self.fdim)
        self.nlat = self.fdim[1]

    lwa_nhem = NHemProperty("lwa", (1, 1))
    lwa_shem = SHemProperty("lwa", (1, 1))


class OutputBarotropicFluxTermsStorage(DerivedQuantityStorage):
    """
    This class stores vertically integrated derived quantities in latitude-longitude 2D grid, including:

    - adv_flux_f1
    - adv_flux_f2
    - adv_flux_f3
    - zonal_adv_flux
    - convergence_zonal_advective_flux
    - divergence_eddy_momentum_flux
    - meridional_heat_flux

    Variables are stored in **python indexing order**: (nlat, nlon)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int,
                 northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        # *** variables below are computed all at once after getting flux terms above?
        self.adv_flux_f1 = np.zeros(self.pydim)
        self.adv_flux_f2 = np.zeros(self.pydim)
        self.adv_flux_f3 = np.zeros(self.pydim)
        self.zonal_adv_flux = np.zeros(self.pydim)
        self.convergence_zonal_advective_flux = np.zeros(self.pydim)
        self.divergence_eddy_momentum_flux = np.zeros(self.pydim)
        self.meridional_heat_flux = np.zeros(self.pydim)


class BarotropicFluxTermsStorage(DerivedQuantityStorage):
    """
    This class stores intermediate computed quantities in latitude-longitude 2D grid.

    Variables are stored in fortran indexing order: (nlon, nlat)
    """
    def __init__(self, pydim: Union[int, Tuple], fdim: Union[int, Tuple],
                 swapaxis_1: int, swapaxis_2: int,
                 northern_hemisphere_results_only: bool):
        super().__init__(pydim, fdim, swapaxis_1, swapaxis_2, northern_hemisphere_results_only)
        self.nlat = fdim[1]
        self.ua1baro = np.zeros(self.fdim)
        self.ua2baro = np.zeros(self.fdim)
        self.ep1baro = np.zeros(self.fdim)
        self.ep2baro = np.zeros(self.fdim)
        self.ep3baro = np.zeros(self.fdim)
        self.ep4 = np.zeros(self.fdim)
        self.u_baro = np.zeros(self.fdim)
        self.lwa_baro = np.zeros(self.fdim)  # This is barotropic LWA (astarbaro)

    ua1baro_nhem = NHemProperty("ua1baro", (1, 0))
    ua1baro_shem = SHemProperty("ua1baro", (1, 0))

    ua2baro_nhem = NHemProperty("ua2baro", (1, 0))
    ua2baro_shem = SHemProperty("ua2baro", (1, 0))

    ep1baro_nhem = NHemProperty("ep1baro", (1, 0))
    ep1baro_shem = SHemProperty("ep1baro", (1, 0))

    ep2baro_nhem = NHemProperty("ep2baro", (1, 0))
    ep2baro_shem = SHemProperty("ep2baro", (1, 0))

    ep3baro_nhem = NHemProperty("ep3baro", (1, 0))
    ep3baro_shem = SHemProperty("ep3baro", (1, 0))

    ep4_nhem = NHemProperty("ep4", (1, 0))
    ep4_shem = SHemProperty("ep4", (1, 0))

    u_baro_nhem = NHemProperty("u_baro", (1, 0))
    u_baro_shem = SHemProperty("u_baro", (1, 0))

    lwa_baro_nhem = NHemProperty("lwa_baro", (1, 0))
    lwa_baro_shem = SHemProperty("lwa_baro", (1, 0))




"""
------------------------------------------
File name: constant.py
Author: Clare Huang
"""
P_GROUND = 1000.  # Ground pressure level. Unit: hPa
SCALE_HEIGHT = 7000.  # Unit: m
CP = 1004.  # specific heat at constant pressure for air (cp) = 1004 J/kg-K
DRY_GAS_CONSTANT = 287.
EARTH_RADIUS = 6.378e+6  # Unit: m
EARTH_OMEGA = 7.29e-5



"""
------------------------------------------
File name: plot_utils.py
Author: Clare Huang
"""
import numpy as np
import xarray as xr
from matplotlib import pyplot as plt  # Optional dependency
from cartopy import crs as ccrs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter


def compare_two_fields_on_map(
    field_a, field_b, a_title, b_title, x_coord, y_coord, title, savefig_fname='default.png',
        diff_factor=0.01, figsize=(15, 4), cmap='rainbow') -> None:

    """
    A handy utility to compare the difference between two 2D-fields and plot their difference. The output plot
    has 3 columns:

    1. value of field_a,

    2. value of field_b,

    3. absolute difference between field_a and field_b.

    The color scale of this plot can be controlled by the parameter *diff_factor*: the color range of the plot will be
    the maximum value among field_a and field_b multiplied by diff_factor. If you want to use the auto-colorscale,
    set diff_factor to None.

    .. versionadded:: 0.7.0

    Parameters
    ----------
        field_a : np.ndarray
            First 2D-field to compare
        field_b : np.ndarray
            Second 2D-field to compare
        a_title :str
            Title of the first field
        b_title :str
            Title of the second field
        x_coord : np.array
            array of x-coordinates
        y_coord : np.array
            array of y-coordinates
        title : str
            Main title of the whole plot
        savefig_fname : str
            Filename of figure saved. Default: "default.png". If you don't want a file to be
            saved, set this to None.
        diff_factor : float
            The color range of the plot will be the maximum value among field_a and field_b
            multiplied by diff_factor. If you want to use the auto-colorscale, set diff_factor to None. Default: 0.01.
        figsize : Tuple[int, int]
            tuple specifying figure size
    """

    cmin = np.min([np.amin(field_a), np.amin(field_b)])
    cmax = np.max([np.amax(field_a), np.amax(field_b)])
    print(f"cmin = {cmin}, cmax = {cmax}")
    fig = plt.figure(figsize=figsize)
    ax1 = fig.add_subplot(1, 3, 1, projection=ccrs.PlateCarree(180))
    cs1 = map_on_axis(ax1, x_coord, y_coord, field_a, levels=np.linspace(cmin, cmax, 31), title=a_title,
                      colorbar_label='')
    ax2 = fig.add_subplot(1, 3, 2, projection=ccrs.PlateCarree(180))
    cs2 = map_on_axis(ax2, x_coord, y_coord, field_b, levels=np.linspace(cmin, cmax, 31), title=b_title,
                      colorbar_label='')
    ax3 = fig.add_subplot(1, 3, 3, projection=ccrs.PlateCarree(180))
    cs3 = map_on_axis(ax3, x_coord, y_coord, np.abs(field_a - field_b),
                      levels=np.linspace(0, diff_factor * max([np.abs(cmin), np.abs(cmax)]), 31),
                      title=f'Abs difference',
                      colorbar_label='')
    plt.suptitle(title)
    plt.tight_layout()
    if savefig_fname:
        plt.savefig(savefig_fname)
    plt.show()


def map_on_axis(ax_obj, xlon, ylat, field, levels, title, colorbar_label='ua2 (m/s**2)'):
    # ax_obj.set_extent([0, 360, ylat.min(), ylat.max()], ccrs.PlateCarree())
    ax_obj.coastlines(color='black', alpha=1)
    ax_obj.set_aspect('auto')
    ax_obj.set_xticks(np.arange(0, 361, 60), crs=ccrs.PlateCarree())
    ax_obj.set_yticks(np.arange(ylat.min(), ylat.max() + 1, 10), crs=ccrs.PlateCarree())
    lon_formatter = LongitudeFormatter(zero_direction_label=True)
    lat_formatter = LatitudeFormatter()
    ax_obj.xaxis.set_major_formatter(lon_formatter)
    ax_obj.yaxis.set_major_formatter(lat_formatter)
    ott = ax_obj.contourf(xlon, ylat, field, levels, transform=ccrs.PlateCarree(), cmap='rainbow')
    ax_obj.set_title(title)
    plt.colorbar(ott, ax=ax_obj, label=colorbar_label)
    return ott


def compare_two_fields(
        field_a, field_b, a_title, b_title, x_coord, y_coord, title, savefig_fname='default.png',
        diff_factor=0.01, figsize=(15, 4), cmap='rainbow') -> None:

    """
    A handy utility to compare the difference between two 2D-fields and plot their difference. The output plot
    has 3 columns:

    1. value of field_a,

    2. value of field_b,

    3. absolute difference between field_a and field_b.

    The color scale of this plot can be controlled by the parameter *diff_factor*: the color range of the plot will be
    the maximum value among field_a and field_b multiplied by diff_factor. If you want to use the auto-colorscale,
    set diff_factor to None.

    .. versionadded:: 0.7.0

    Parameters
    ----------
        field_a : np.ndarray
            First 2D-field to compare
        field_b : np.ndarray
            Second 2D-field to compare
        a_title :str
            Title of the first field
        b_title :str
            Title of the second field
        x_coord : np.array
            array of x-coordinates
        y_coord : np.array
            array of y-coordinates
        title : str
            Main title of the whole plot
        savefig_fname : str
            Filename of figure saved. Default: "default.png". If you don't want a file to be
            saved, set this to None.
        diff_factor : float
            The color range of the plot will be the maximum value among field_a and field_b
            multiplied by diff_factor. If you want to use the auto-colorscale, set diff_factor to None. Default: 0.01.
        figsize : Tuple[int, int]
            tuple specifying figure size
    """

    cmin = np.min([np.amin(field_a), np.amin(field_b)])
    cmax = np.max([np.amax(field_a), np.amax(field_b)])
    print(f"cmin = {cmin}, cmax = {cmax}")
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=figsize)
    cs1 = ax1.contourf(x_coord, y_coord, field_a, np.linspace(cmin, cmax, 31), cmap=cmap)
    ax1.set_title(a_title)
    cbar1 = plt.colorbar(cs1, ax=ax1)
    cs2 = ax2.contourf(x_coord, y_coord, field_b, np.linspace(cmin, cmax, 31), cmap=cmap)
    ax2.set_title(b_title)
    cbar2 = plt.colorbar(cs2, ax=ax2)
    if diff_factor:
        cs3 = ax3.contourf(x_coord, y_coord, np.abs(field_a-field_b),
                           np.linspace(0, diff_factor * max([np.abs(cmin), np.abs(cmax)]), 31), cmap=cmap)
    else:
        cs3 = ax3.contourf(x_coord, y_coord, np.abs(field_a - field_b), cmap=cmap)
    ax3.set_title(f'Abs difference')
    cbar3 = plt.colorbar(cs3, ax=ax3)
    plt.suptitle(title)
    plt.tight_layout()
    if savefig_fname:
        plt.savefig(savefig_fname)
    plt.show()


def plot_lon_lat_field(filepath, variable_name, latitude_name='latitude', longitude_name='longitude',
                       tstep=10, zonal_axis=0) -> None:
    """
    Plot a snapshot of longitude-latitude map from a netCDF file.

    Parameters
    ----------
        filepath : str
            path to the netCDF file
        variable_name : str
            name of the variable to be plotted
        latitude_name : str
            name of latitudinal coordinates
        longitude_name : str
            name of latitudinal coordinates
        tstep : int
            index of timestep to be plotted
        zonal_axis : int
            axis of zonal dimension
    """
    file_handle = xr.open_dataset(filepath)
    field = file_handle.variables[variable_name].isel(time=tstep).values
    print(f"Zonal mean of the field:\n{field.mean(axis=zonal_axis)}.")

    # plot
    plt.contourf(
        file_handle.coords[longitude_name],
        file_handle.coords[latitude_name],
        file_handle.variables[variable_name].isel(time=tstep))
    plt.show()
